import os
import signal
import time
import copy
from datetime import datetime, timedelta
from typing import Dict, List, Set, Callable, Literal, Optional, Any
from decimal import Decimal
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from collections import defaultdict
from nexustrader.base import ExchangeManager
from nexustrader.indicator import IndicatorManager, Indicator, IndicatorProxy
from nexustrader.core.entity import TaskManager, DataReady, is_redis_available
from nexustrader.core.cache import AsyncCache
from nexustrader.error import StrategyBuildError
from nexustrader.base import (
    ExecutionManagementSystem,
    PrivateConnector,
    PublicConnector,
)
from nexustrader.core.entity import OidGen
from nexustrader.core.nautilius_core import MessageBus, LiveClock, Logger
from nexustrader.schema import (
    BookL1,
    Trade,
    Kline,
    BookL2,
    Order,
    FundingRate,
    Ticker,
    IndexPrice,
    MarkPrice,
    InstrumentId,
    BaseMarket,
    AccountBalance,
    CreateOrderSubmit,
    TakeProfitAndStopLossOrderSubmit,
    # TWAPOrderSubmit,
    ModifyOrderSubmit,
    CancelOrderSubmit,
    CancelAllOrderSubmit,
    # CancelTWAPOrderSubmit,
    KlineList,
    BatchOrder,
    BatchOrderSubmit,
)
from nexustrader.constants import (
    DataType,
    BookLevel,
    OrderSide,
    OrderType,
    TimeInForce,
    # PositionSide,
    AccountType,
    SubmitType,
    ExchangeType,
    KlineInterval,
    TriggerType,
)


class Strategy:
    def __init__(self):
        self.log = Logger(name=type(self).__name__)

        self._subscriptions: Dict[
            DataType,
            Dict[KlineInterval, Set[str]] | Set[str] | Dict[BookLevel, Set[str]],
        ] = {
            DataType.BOOKL1: set(),
            DataType.BOOKL2: defaultdict(set),
            DataType.TRADE: set(),
            DataType.KLINE: defaultdict(set),
            DataType.FUNDING_RATE: set(),
            DataType.INDEX_PRICE: set(),
            DataType.MARK_PRICE: set(),
        }

        self._initialized = False
        self._scheduler = AsyncIOScheduler()
        self.indicator = IndicatorProxy()

    def _init_core(
        self,
        exchanges: Dict[ExchangeType, ExchangeManager],
        public_connectors: Dict[AccountType, PublicConnector],
        private_connectors: Dict[AccountType, PrivateConnector],
        cache: AsyncCache,
        msgbus: MessageBus,
        clock: LiveClock,
        task_manager: TaskManager,
        ems: Dict[ExchangeType, ExecutionManagementSystem],
        strategy_id: str = None,
        user_id: str = None,
        enable_cli: bool = False,
    ):
        if self._initialized:
            return

        self.cache = cache
        self.clock = clock
        self._oidgen = OidGen(clock)
        self._ems = ems
        self._task_manager = task_manager
        self._msgbus = msgbus
        self._private_connectors = private_connectors
        self._public_connectors = public_connectors
        self._exchanges = exchanges
        self._indicator_manager = IndicatorManager(self._msgbus)

        # Initialize state exporter if IDs are provided and Redis is fully available
        self._state_exporter = None
        if strategy_id and user_id and is_redis_available() and enable_cli:
            try:
                from nexustrader.cli.monitor.state_exporter import StrategyStateExporter

                self._state_exporter = StrategyStateExporter(
                    strategy_id=strategy_id, user_id=user_id, cache=cache, clock=clock
                )
                self.log.debug("CLI monitoring enabled with Redis")
            except Exception as e:
                self.log.debug(
                    f"State exporter initialization failed, CLI monitoring disabled: {e}"
                )
        elif strategy_id and user_id:
            self.log.debug("Redis not available, CLI monitoring disabled")

        self._msgbus.register(endpoint="pending", handler=self.on_pending_order)
        self._msgbus.register(endpoint="accepted", handler=self.on_accepted_order)
        self._msgbus.register(
            endpoint="partially_filled", handler=self.on_partially_filled_order
        )
        self._msgbus.register(endpoint="filled", handler=self.on_filled_order)
        self._msgbus.register(endpoint="canceling", handler=self.on_canceling_order)
        self._msgbus.register(endpoint="canceled", handler=self.on_canceled_order)
        self._msgbus.register(endpoint="failed", handler=self.on_failed_order)
        self._msgbus.register(
            endpoint="cancel_failed", handler=self.on_cancel_failed_order
        )

        self._msgbus.register(endpoint="balance", handler=self.on_balance)

        self._initialized = True

        self._subscriptions_ready: Dict[DataType | str, DataReady] = {}

    @property
    def ready(self):
        return all(
            data_ready.ready for data_ready in self._subscriptions_ready.values()
        )

    def api(self, account_type: AccountType):
        return self._private_connectors[account_type].api

    def register_indicator(
        self,
        symbols: str | List[str],
        indicator: Indicator,
        data_type: DataType,
        account_type: AccountType | None = None,
    ):
        if not self._initialized:
            raise StrategyBuildError(
                "Strategy not initialized, please use `register_indicator` in `on_start` method"
            )

        if isinstance(symbols, str):
            symbols = [symbols]

        # Create separate indicator instances for each symbol to avoid shared state
        for symbol in symbols:
            # Create a deep copy of the indicator for each symbol
            symbol_indicator = copy.deepcopy(indicator)

            # Register the symbol-specific indicator with the proxy
            self.indicator.register_indicator(indicator.name, symbol, symbol_indicator)

            match data_type:
                case DataType.BOOKL1:
                    self._indicator_manager.add_bookl1_indicator(
                        symbol, symbol_indicator
                    )
                case DataType.BOOKL2:
                    self._indicator_manager.add_bookl2_indicator(
                        symbol, symbol_indicator
                    )
                case DataType.KLINE:
                    self._indicator_manager.add_kline_indicator(
                        symbol, symbol_indicator
                    )
                    # Handle warmup for kline indicators
                    if symbol_indicator.requires_warmup:
                        if not account_type:
                            # Infer account type if not provided
                            account_type = self._infer_account_type(symbol)
                        self._perform_indicator_warmup(
                            symbol, symbol_indicator, account_type
                        )
                case DataType.TRADE:
                    self._indicator_manager.add_trade_indicator(
                        symbol, symbol_indicator
                    )
                case DataType.INDEX_PRICE:
                    self._indicator_manager.add_index_price_indicator(
                        symbol, symbol_indicator
                    )
                case DataType.FUNDING_RATE:
                    self._indicator_manager.add_funding_rate_indicator(
                        symbol, symbol_indicator
                    )
                case DataType.MARK_PRICE:
                    self._indicator_manager.add_mark_price_indicator(
                        symbol, symbol_indicator
                    )
                case _:
                    raise ValueError(f"Invalid data type: {data_type}")

    def _infer_account_type(self, symbol: str) -> AccountType:
        """
        Infer the account type based on the symbol's exchange and type.
        This is useful for methods that require an account type but don't have it explicitly provided.
        """
        instrument_id = InstrumentId.from_str(symbol)
        exchange = self._exchanges.get(instrument_id.exchange)
        if not exchange:
            raise ValueError(
                f"Exchange {instrument_id.exchange} not found, please add it to the config"
            )
        return exchange.instrument_id_to_account_type(instrument_id)

    def request_ticker(
        self,
        symbol: str,
        account_type: AccountType | None = None,
    ) -> Ticker:
        account_type = account_type or self._infer_account_type(symbol)
        connector = self._public_connectors.get(account_type)
        if not connector:
            raise ValueError(
                f"Account type {account_type} not found in public connectors"
            )
        return connector.request_ticker(symbol)

    def request_all_tickers(
        self,
        account_type: AccountType,
    ) -> Dict[str, Ticker]:
        connector = self._public_connectors.get(account_type)
        if not connector:
            raise ValueError(
                f"Account type {account_type} not found in public connectors"
            )
        return connector.request_all_tickers()

    def request_klines(
        self,
        symbol: str | List[str],
        interval: KlineInterval,
        limit: int | None = None,
        start_time: int | datetime | None = None,
        end_time: int | datetime | None = None,
        account_type: AccountType | None = None,
    ) -> KlineList:
        if isinstance(start_time, datetime):
            start_time = int(start_time.timestamp() * 1000)
        if isinstance(end_time, datetime):
            end_time = int(end_time.timestamp() * 1000)

        account_type = account_type or self._infer_account_type(symbol)
        connector = self._public_connectors.get(account_type)
        if not connector:
            raise ValueError(
                f"Account type {account_type} not found in public connectors"
            )

        if isinstance(symbol, str):
            symbol = [symbol]

        klines = KlineList([])
        for sym in symbol:
            res = connector.request_klines(
                symbol=sym,
                interval=interval,
                limit=limit,
                start_time=start_time,
                end_time=end_time,
            )
            klines.extend(res)
        return klines

    def request_index_klines(
        self,
        symbol: str | List[str],
        interval: KlineInterval,
        limit: int | None = None,
        start_time: int | datetime | None = None,
        end_time: int | datetime | None = None,
        account_type: AccountType | None = None,
    ) -> KlineList:
        if isinstance(start_time, datetime):
            start_time = int(start_time.timestamp() * 1000)
        if isinstance(end_time, datetime):
            end_time = int(end_time.timestamp() * 1000)
        account_type = account_type or self._infer_account_type(symbol)
        connector = self._public_connectors.get(account_type)
        if not connector:
            raise ValueError(
                f"Account type {account_type} not found in public connectors"
            )

        if isinstance(symbol, str):
            symbol = [symbol]

        klines = KlineList([])
        for sym in symbol:
            res = connector.request_index_klines(
                symbol=sym,
                interval=interval,
                limit=limit,
                start_time=start_time,
                end_time=end_time,
            )
            klines.extend(res)
        return klines

    def _perform_indicator_warmup(
        self, symbol: str, indicator: Indicator, account_type: AccountType
    ):
        """Automatically fetch historical data to warm up an indicator."""
        try:
            # Calculate how much historical data we need
            warmup_microseconds = (
                indicator.warmup_period * indicator.warmup_interval.microseconds
            )
            start_time_ms = self.clock.timestamp_ms() - warmup_microseconds

            # Fetch historical klines
            historical_klines = self.request_klines(
                symbol=symbol,
                account_type=account_type,
                interval=indicator.warmup_interval,
                limit=indicator.warmup_period,
                start_time=start_time_ms,
            )

            # Process historical data for warmup (oldest first)
            for kline in historical_klines.values:
                if kline.symbol == symbol and kline.confirm:
                    indicator._process_warmup_kline(kline)

            self.log.debug(
                f"Warmed up indicator {indicator.name} for {symbol} with {len(historical_klines)} klines"
            )

        except Exception as e:
            self.log.error(
                f"Failed to warm up indicator {indicator.name} for {symbol}: {e}"
            )

    def get_warmup_status(self) -> dict[str, list[dict]]:
        """Get the warmup status of all indicators by symbol."""
        status = {}
        requirements = self._indicator_manager.get_warmup_requirements()

        for symbol, indicator_list in requirements.items():
            status[symbol] = []
            for indicator, period, interval in indicator_list:
                status[symbol].append(
                    {
                        "name": indicator.name,
                        "warmup_period": period,
                        "warmup_interval": interval.value,
                        "is_warmed_up": indicator.is_warmed_up,
                        "data_count": indicator._warmup_data_count,
                    }
                )

        return status

    def wait_for_warmup(self, timeout_seconds: int = 60) -> bool:
        """Wait for all indicators to complete warmup. Returns True if all warmed up."""
        start_time = self.clock.timestamp()
        while self.clock.timestamp() - start_time < timeout_seconds:
            if not self._indicator_manager.has_warmup_pending():
                return True

        return False

    def schedule(
        self,
        func: Callable,
        trigger: Literal["interval", "cron", "date"] = "interval",
        **kwargs,
    ):
        """
        There are three modes:

        - **cron**: run at a specific time second, minute, hour, day, month, year
        - **interval**: run at a specific interval  seconds, minutes, hours, days, weeks, months, years
        - **date**: run at a specific date and time, `run_date` must be provided

        kwargs:
            next_run_time: datetime, when to run the first time
            seconds/minutes/hours/days/weeks: int, interval between runs
            year/month/day/hour/minute/second: int, specific time to run
        """
        if not self._initialized:
            raise RuntimeError(
                "Strategy not initialized, please use `schedule` in `on_start` method"
            )
        self._scheduler.add_job(func, trigger=trigger, **kwargs)

    def market(self, symbol: str) -> BaseMarket:
        instrument_id = InstrumentId.from_str(symbol)
        exchange = self._exchanges[instrument_id.exchange]
        return exchange.market[instrument_id.symbol]

    def min_order_amount(self, symbol: str) -> Decimal:
        instrument_id = InstrumentId.from_str(symbol)
        ems = self._ems[instrument_id.exchange]
        return ems._get_min_order_amount(instrument_id.symbol, self.market(symbol))

    def amount_to_precision(
        self,
        symbol: str,
        amount: float,
        mode: Literal["round", "ceil", "floor"] = "round",
    ) -> Decimal:
        instrument_id = InstrumentId.from_str(symbol)
        ems = self._ems[instrument_id.exchange]
        return ems._amount_to_precision(instrument_id.symbol, amount, mode)

    def price_to_precision(
        self,
        symbol: str,
        price: float,
        mode: Literal["round", "ceil", "floor"] = "round",
    ) -> Decimal:
        instrument_id = InstrumentId.from_str(symbol)
        ems = self._ems[instrument_id.exchange]
        return ems._price_to_precision(instrument_id.symbol, price, mode)

    def create_batch_orders(
        self,
        orders: List[BatchOrder],
        account_type: AccountType | None = None,
    ):
        """
        Create a batch of orders.

        Args:
            orders (List[BatchOrder]): A list of BatchOrder objects to be submitted.
            account_type (AccountType | None): The account type for the orders. If None, it will auto selected the account_type, but for performance issue, recommend to set.
        """
        batch_orders: list[BatchOrderSubmit] = []
        for order in orders:
            batch_order = BatchOrderSubmit(
                symbol=order.symbol,
                instrument_id=InstrumentId.from_str(order.symbol),
                side=order.side,
                type=order.type,
                oid=self._oidgen.oid,
                amount=order.amount,
                price=order.price,
                time_in_force=order.time_in_force,
                reduce_only=order.reduce_only,
                kwargs=order.kwargs,
            )
            batch_orders.append(batch_order)
        self._ems[batch_orders[0].instrument_id.exchange]._submit_order(
            batch_orders, SubmitType.BATCH, account_type
        )
        return [order.oid for order in batch_orders]

    def create_tp_sl_order(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce | None = TimeInForce.GTC,
        tp_order_type: OrderType | None = None,
        tp_trigger_price: Decimal | None = None,
        tp_price: Decimal | None = None,
        tp_trigger_type: TriggerType = TriggerType.LAST_PRICE,
        sl_order_type: OrderType | None = None,
        sl_trigger_price: Decimal | None = None,
        sl_price: Decimal | None = None,
        sl_trigger_type: TriggerType = TriggerType.LAST_PRICE,
        account_type: AccountType | None = None,
        **kwargs,
    ):
        order = TakeProfitAndStopLossOrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
            side=side,
            type=type,
            oid=self._oidgen.oid,
            amount=amount,
            price=price,
            time_in_force=time_in_force,
            tp_order_type=tp_order_type,
            tp_trigger_price=tp_trigger_price,
            tp_price=tp_price,
            tp_trigger_type=tp_trigger_type,
            sl_order_type=sl_order_type,
            sl_trigger_price=sl_trigger_price,
            sl_price=sl_price,
            sl_trigger_type=sl_trigger_type,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(
            order, SubmitType.TAKE_PROFIT_AND_STOP_LOSS, account_type
        )
        return order.oid

    def create_order(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce | None = TimeInForce.GTC,
        reduce_only: bool = False,
        account_type: AccountType | None = None,
        **kwargs,
    ) -> str:
        order = CreateOrderSubmit(
            symbol=symbol,
            oid=self._oidgen.oid,
            instrument_id=InstrumentId.from_str(symbol),
            side=side,
            type=type,
            amount=amount,
            price=price,
            time_in_force=time_in_force,
            reduce_only=reduce_only,
            # position_side=position_side,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(
            order, SubmitType.CREATE, account_type
        )
        return order.oid

    def create_order_ws(
        self,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce | None = TimeInForce.GTC,
        reduce_only: bool = False,
        account_type: AccountType | None = None,
        **kwargs,
    ) -> str:
        order = CreateOrderSubmit(
            symbol=symbol,
            oid=self._oidgen.oid,
            instrument_id=InstrumentId.from_str(symbol),
            side=side,
            type=type,
            amount=amount,
            price=price,
            time_in_force=time_in_force,
            reduce_only=reduce_only,
            # position_side=position_side,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(
            order, SubmitType.CREATE_WS, account_type
        )
        return order.oid

    def cancel_order(
        self, symbol: str, oid: str, account_type: AccountType | None = None, **kwargs
    ) -> str:
        order = CancelOrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
            oid=oid,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(
            order, SubmitType.CANCEL, account_type
        )
        return order.oid

    def cancel_order_ws(
        self, symbol: str, oid: str, account_type: AccountType | None = None, **kwargs
    ) -> str:
        order = CancelOrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
            oid=oid,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(
            order, SubmitType.CANCEL_WS, account_type
        )
        return order.oid

    def cancel_all_orders(
        self, symbol: str, account_type: AccountType | None = None
    ) -> str:
        order = CancelAllOrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
        )
        self._ems[order.instrument_id.exchange]._submit_order(
            order, SubmitType.CANCEL_ALL, account_type
        )

    def modify_order(
        self,
        symbol: str,
        oid: str,
        side: OrderSide | None = None,
        price: Decimal | None = None,
        amount: Decimal | None = None,
        account_type: AccountType | None = None,
        **kwargs,
    ) -> str:
        order = ModifyOrderSubmit(
            symbol=symbol,
            instrument_id=InstrumentId.from_str(symbol),
            oid=oid,
            side=side,
            price=price,
            amount=amount,
            kwargs=kwargs,
        )
        self._ems[order.instrument_id.exchange]._submit_order(
            order, SubmitType.MODIFY, account_type
        )
        return order.oid

    # def create_twap(
    #     self,
    #     symbol: str,
    #     side: OrderSide,
    #     amount: Decimal,
    #     duration: int,
    #     wait: int,
    #     check_interval: float = 0.1,
    #     position_side: PositionSide | None = None,
    #     account_type: AccountType | None = None,
    #     **kwargs,
    # ) -> str:
    #     order = TWAPOrderSubmit(
    #         symbol=symbol,
    #         instrument_id=InstrumentId.from_str(symbol),
    #         side=side,
    #         amount=amount,
    #         duration=duration,
    #         wait=wait,
    #         check_interval=check_interval,
    #         position_side=position_side,
    #         kwargs=kwargs,
    #     )
    #     self._ems[order.instrument_id.exchange]._submit_order(
    #         order, SubmitType.TWAP, account_type
    #     )
    #     return order.uuid

    # def cancel_twap(
    #     self, symbol: str, uuid: str, account_type: AccountType | None = None
    # ) -> str:
    #     order = CancelTWAPOrderSubmit(
    #         symbol=symbol,
    #         instrument_id=InstrumentId.from_str(symbol),
    #         uuid=uuid,
    #     )
    #     self._ems[order.instrument_id.exchange]._submit_order(
    #         order, SubmitType.CANCEL_TWAP, account_type
    #     )
    #     return order.uuid

    def subscribe_bookl1(
        self, symbols: str | List[str], ready_timeout: int = 60, ready: bool = True
    ):
        """
        Subscribe to level 1 book data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
            ready_timeout (int): The timeout for the data to be ready.
            ready (bool): default is True. Whether the data is ready. If True, the data will be ready immediately. When you use event driven strategy, you can set it to True. Otherwise, set it to False.
        """
        if not self._initialized:
            raise StrategyBuildError(
                "Strategy not initialized, please use `subscribe_bookl1` in `on_start` method"
            )

        self._msgbus.subscribe(topic="bookl1", handler=self._on_bookl1)
        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            self._subscriptions[DataType.BOOKL1].add(symbol)

        self._subscriptions_ready[DataType.BOOKL1] = DataReady(
            symbols,
            name="bookl1",
            timeout=ready_timeout,
            permanently_ready=ready,
        )

    def subscribe_trade(
        self, symbols: str | List[str], ready_timeout: int = 60, ready: bool = True
    ):
        """
        Subscribe to trade data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
            ready_timeout (int): The timeout for the data to be ready.
            ready (bool): default is True. Whether the data is ready. If True, the data will be ready immediately. When you use event driven strategy, you can set it to True. Otherwise, set it to False.
        """
        if not self._initialized:
            raise StrategyBuildError(
                "Strategy not initialized, please use `subscribe_trade` in `on_start` method"
            )

        self._msgbus.subscribe(topic="trade", handler=self._on_trade)

        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            self._subscriptions[DataType.TRADE].add(symbol)

        self._subscriptions_ready[DataType.TRADE] = DataReady(
            symbols,
            name="trade",
            timeout=ready_timeout,
            permanently_ready=ready,
        )

    def subscribe_kline(
        self,
        symbols: str | List[str],
        interval: KlineInterval,
        ready_timeout: int = 60,
        ready: bool = True,
    ):
        """
        Subscribe to kline data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
            interval (str): The interval of the kline data
            ready_timeout (int): The timeout for the data to be ready.
            ready (bool): default is True. Whether the data is ready. If True, the data will be ready immediately. When you use event driven strategy, you can set it to True. Otherwise, set it to False.
        """
        if not self._initialized:
            raise StrategyBuildError(
                "Strategy not initialized, please use `subscribe_kline` in `on_start` method"
            )

        self._msgbus.subscribe(topic="kline", handler=self._on_kline)

        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            self._subscriptions[DataType.KLINE][interval].add(symbol)

        self._subscriptions_ready[interval.value] = DataReady(
            symbols,
            name=f"kline_{interval.value}",
            timeout=ready_timeout,
            permanently_ready=ready,
        )

    def subscribe_bookl2(
        self,
        symbols: str | List[str],
        level: BookLevel,
        ready_timeout: int = 60,
        ready: bool = True,
    ):
        """
        Subscribe to level 2 book data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
            level (BookLevel): The level of the book data
            ready_timeout (int): The timeout for the data to be ready.
            ready (bool): default is True. Whether the data is ready. If True, the data will be ready immediately. When you use event driven strategy, you can set it to True. Otherwise, set it to False.
        """
        if not self._initialized:
            raise StrategyBuildError(
                "Strategy not initialized, please use `subscribe_bookl2` in `on_start` method"
            )

        self._msgbus.subscribe(topic="bookl2", handler=self._on_bookl2)

        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            self._subscriptions[DataType.BOOKL2][level].add(symbol)

        self._subscriptions_ready[DataType.BOOKL2] = DataReady(
            symbols,
            name="bookl2",
            timeout=ready_timeout,
            permanently_ready=ready,
        )

    def subscribe_funding_rate(
        self, symbols: str | List[str], ready_timeout: int = 60, ready: bool = True
    ):
        """
        Subscribe to funding rate data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
            ready_timeout (int): The timeout for the data to be ready.
            ready (bool): default is True. Whether the data is ready. If True, the data will be ready immediately. When you use event driven strategy, you can set it to True. Otherwise, set it to False.
        """
        if not self._initialized:
            raise StrategyBuildError(
                "Strategy not initialized, please use `subscribe_funding_rate` in `on_start` method"
            )

        self._msgbus.subscribe(topic="funding_rate", handler=self._on_funding_rate)

        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            self._subscriptions[DataType.FUNDING_RATE].add(symbol)

        self._subscriptions_ready[DataType.FUNDING_RATE] = DataReady(
            symbols,
            name="funding_rate",
            timeout=ready_timeout,
            permanently_ready=ready,
        )

    def subscribe_index_price(
        self, symbols: str | List[str], ready_timeout: int = 60, ready: bool = True
    ):
        """
        Subscribe to index price data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
            ready_timeout (int): The timeout for the data to be ready.
            ready (bool): default is True. Whether the data is ready. If True, the data will be ready immediately. When you use event driven strategy, you can set it to True. Otherwise, set it to False.
        """
        if not self._initialized:
            raise StrategyBuildError(
                "Strategy not initialized, please use `subscribe_index_price` in `on_start` method"
            )

        self._msgbus.subscribe(topic="index_price", handler=self._on_index_price)

        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            self._subscriptions[DataType.INDEX_PRICE].add(symbol)

        self._subscriptions_ready[DataType.INDEX_PRICE] = DataReady(
            symbols,
            name="index_price",
            timeout=ready_timeout,
            permanently_ready=ready,
        )

    def subscribe_mark_price(
        self, symbols: str | List[str], ready_timeout: int = 60, ready: bool = True
    ):
        """
        Subscribe to mark price data for the given symbols.

        Args:
            symbols (List[str]): The symbols to subscribe to.
            ready_timeout (int): The timeout for the data to be ready.
            ready (bool): default is True. Whether the data is ready. If True, the data will be ready immediately. When you use event driven strategy, you can set it to True. Otherwise, set it to False.
        """
        if not self._initialized:
            raise StrategyBuildError(
                "Strategy not initialized, please use `subscribe_mark_price` in `on_start` method"
            )

        self._msgbus.subscribe(topic="mark_price", handler=self._on_mark_price)

        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            self._subscriptions[DataType.MARK_PRICE].add(symbol)

        self._subscriptions_ready[DataType.MARK_PRICE] = DataReady(
            symbols,
            name="mark_price",
            timeout=ready_timeout,
            permanently_ready=ready,
        )

    def linear_info(
        self,
        exchange: ExchangeType,
        base: str | None = None,
        quote: str | None = None,
        exclude: List[str] | None = None,
    ) -> List[str]:
        exchange: ExchangeManager = self._exchanges[exchange]
        return exchange.linear(base, quote, exclude)

    def spot_info(
        self,
        exchange: ExchangeType,
        base: str | None = None,
        quote: str | None = None,
        exclude: List[str] | None = None,
    ) -> List[str]:
        exchange: ExchangeManager = self._exchanges[exchange]
        return exchange.spot(base, quote, exclude)

    def future_info(
        self,
        exchange: ExchangeType,
        base: str | None = None,
        quote: str | None = None,
        exclude: List[str] | None = None,
    ) -> List[str]:
        exchange: ExchangeManager = self._exchanges[exchange]
        return exchange.future(base, quote, exclude)

    def inverse_info(
        self,
        exchange: ExchangeType,
        base: str | None = None,
        quote: str | None = None,
        exclude: List[str] | None = None,
    ) -> List[str]:
        exchange: ExchangeManager = self._exchanges[exchange]
        return exchange.inverse(base, quote, exclude)

    def on_start(self):
        pass

    def on_stop(self):
        pass

    def _on_start(self):
        # Start state exporter if available
        if self._state_exporter:
            self._state_exporter.start()
        self.on_start()

    def _on_stop(self):
        # Stop state exporter if available
        if self._state_exporter:
            self._state_exporter.stop()
        self.on_stop()

    def on_trade(self, trade: Trade):
        pass

    def on_bookl1(self, bookl1: BookL1):
        pass

    def on_bookl2(self, bookl2: BookL2):
        pass

    def on_kline(self, kline: Kline):
        pass

    def on_funding_rate(self, funding_rate: FundingRate):
        pass

    def on_index_price(self, index_price: IndexPrice):
        pass

    def on_mark_price(self, mark_price: MarkPrice):
        pass

    def on_pending_order(self, order: Order):
        pass

    def on_accepted_order(self, order: Order):
        pass

    def on_partially_filled_order(self, order: Order):
        pass

    def on_filled_order(self, order: Order):
        pass

    def on_canceling_order(self, order: Order):
        pass

    def on_canceled_order(self, order: Order):
        pass

    def on_failed_order(self, order: Order):
        pass

    def on_cancel_failed_order(self, order: Order):
        pass

    def on_balance(self, balance: AccountBalance):
        pass

    def stop(self):
        time.sleep(0.2)  # wait for 200ms to ensure all messages are processed
        os.kill(os.getpid(), signal.SIGINT)

    def wait(self, seconds: int):
        time.sleep(seconds)

    def _on_trade(self, trade: Trade):
        self.on_trade(trade)
        self._subscriptions_ready[DataType.TRADE].input(trade)

    def _on_bookl2(self, bookl2: BookL2):
        self.on_bookl2(bookl2)
        self._subscriptions_ready[DataType.BOOKL2].input(bookl2)

    def _on_kline(self, kline: Kline):
        self.on_kline(kline)
        self._subscriptions_ready[kline.interval.value].input(kline)

    def _on_funding_rate(self, funding_rate: FundingRate):
        self.on_funding_rate(funding_rate)
        self._subscriptions_ready[DataType.FUNDING_RATE].input(funding_rate)

    def _on_bookl1(self, bookl1: BookL1):
        self.on_bookl1(bookl1)
        self._subscriptions_ready[DataType.BOOKL1].input(bookl1)

    def _on_index_price(self, index_price: IndexPrice):
        self.on_index_price(index_price)
        self._subscriptions_ready[DataType.INDEX_PRICE].input(index_price)

    def _on_mark_price(self, mark_price: MarkPrice):
        self.on_mark_price(mark_price)
        self._subscriptions_ready[DataType.MARK_PRICE].input(mark_price)

    def param(
        self, name: str, value: Optional[Any] = None, default: Optional[Any] = None
    ) -> Any:
        """
        Get or set a parameter in the cache.

        Args:
            name: The parameter name
            value: The parameter value to set. If None, will get the parameter.

        Returns:
            The parameter value if getting, None if setting.

        Examples:
            # Set a parameter
            self.param('rolling_n', 10)

            # Get a parameter
            rolling_n = self.param('rolling_n')
        """
        if value is not None:
            # Set parameter
            self.cache.set_param(name, value)
            return None
        else:
            # Get parameter
            return self.cache.get_param(name, default)

    def clear_param(self, name: Optional[str] = None) -> None:
        """
        Clear parameter(s) from the cache.

        Args:
            name: The parameter name to clear. If None, clears all parameters.

        Examples:
            # Clear a specific parameter
            self.clear_param('rolling_n')

            # Clear all parameters
            self.clear_param()
        """
        self.cache.clear_param(name)

    def set_timer(
        self,
        callback: Callable,
        interval: timedelta,
        name: str | None = None,
        start_time: datetime | None = None,
        stop_time: datetime | None = None,
    ):
        """
        Set a timer that calls a callback function at regular intervals.

        Args:
            callback: The function to call
            interval: Time interval between calls
            name: Optional timer name. If not provided, uses the callback function name.
            start_time: When to start the timer (defaults to now + interval)
            stop_time: When to stop the timer (optional)
        """
        if name is None:
            name = callback.__name__

        if start_time is None:
            start_time = self.clock.utc_now() + interval

        self.clock.set_timer(
            name=name,
            interval=interval,
            start_time=start_time,
            stop_time=stop_time,
            callback=callback,
        )



from typing import Dict, Any
import msgspec
import redis
from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, TabbedContent, TabPane, Select
from textual.containers import Container

from nexustrader.cli.widgets.dashboard import DashboardWidget
from nexustrader.cli.widgets.positions import PositionsWidget
from nexustrader.cli.widgets.balances import BalancesWidget
from nexustrader.core.entity import get_redis_client_if_available


class NexusTraderMonitor(App):
    """NexusTrader CLI monitoring application"""

    def __init__(self):
        super().__init__()
        self.current_strategy = None
        self.strategies = {}
        self._redis_client = None
        self._init_redis()

    CSS = """
    Screen {
        background: $surface;
        layout: vertical;
    }
    
    .widget {
        border: solid $primary;
        height: auto;
        margin: 1;
    }
    
    .strategy-selector {
        height: 4;
        margin: 1 2;
        dock: top;
    }
    
    Select {
        width: 100%;
        height: 100%;
    }
    
    TabbedContent {
        height: 1fr;
        margin-top: 1;
    }
    
    TabPane {
        padding: 1;
    }
    """

    def compose(self) -> ComposeResult:
        yield Header()
        with Container(classes="strategy-selector"):
            yield Select(
                [], prompt="Select a strategy to monitor", id="strategy-select"
            )
        with TabbedContent():
            with TabPane("📊 Dashboard"):
                yield DashboardWidget()
            with TabPane("📍 Positions"):
                yield PositionsWidget()
            with TabPane("💰 Balances"):
                yield BalancesWidget()
        yield Footer()

    @on(Select.Changed)
    def select_changed(self, event: Select.Changed) -> None:
        if event.value != Select.BLANK:
            self.current_strategy = str(event.value)
            self.title = f"NexusTrader Monitor - {self.current_strategy}"

    def on_mount(self):
        self.title = "NexusTrader Monitor"
        self.set_interval(1.0, self.refresh_strategies)
        self.set_interval(1.0, self.refresh_data)

    def refresh_strategies(self) -> None:
        """Refresh the list of running strategies from Redis"""
        if not self._redis_client:
            return

        try:
            # Get all running strategies from Redis hash
            running_strategies = self._redis_client.hgetall(
                "nexustrader:running_strategies"
            )

            strategies = {}
            for strategy_key, strategy_data in running_strategies.items():
                if isinstance(strategy_key, bytes):
                    strategy_key = strategy_key.decode("utf-8")
                if isinstance(strategy_data, bytes):
                    strategy_data = strategy_data.decode("utf-8")

                try:
                    strategy_info = msgspec.json.decode(strategy_data)
                    full_id = strategy_info.get("full_id", strategy_key)
                    pid = strategy_info.get("pid", "unknown")
                    strategies[f"{full_id}:{pid}"] = strategy_info
                except msgspec.DecodeError:
                    continue

            self.strategies = strategies
            self._update_strategy_select()

            # Auto-select first strategy if none selected
            if not self.current_strategy and self.strategies:
                first_strategy_id = next(iter(self.strategies.keys()))
                self.current_strategy = first_strategy_id
                self.title = f"NexusTrader Monitor - {self.current_strategy}"
        except Exception:
            # Redis might not be available yet
            pass

    def refresh_data(self) -> None:
        """Refresh data for the selected strategy"""
        if not self.current_strategy:
            return

        try:
            strategy_data = self._load_strategy_data(self.current_strategy)
            self._update_widgets(strategy_data)
        except Exception:
            # Data might not be available yet
            pass

    def _update_strategy_select(self) -> None:
        """Update the strategy selector options"""
        select = self.query_one("#strategy-select", Select)
        options = [(strategy_id, strategy_id) for strategy_id in self.strategies.keys()]
        current_value = select.value
        select.set_options(options)
        # Restore selection if it still exists
        if current_value and current_value in self.strategies:
            select.value = current_value

    def _init_redis(self):
        """Initialize Redis connection"""
        self._redis_client = get_redis_client_if_available()

    def _load_strategy_data(self, strategy_id: str) -> Dict[str, Any]:
        """Load strategy data from Redis"""
        if not self._redis_client:
            return {}

        data = {}

        # Load each data type from Redis
        for data_type in ["positions", "balances", "metrics"]:
            try:
                key = f"strategy:{strategy_id}:{data_type}"
                redis_data = self._redis_client.get(key)
                if redis_data:
                    content = msgspec.json.decode(redis_data)
                    data[data_type] = content
            except (redis.RedisError, redis.ConnectionError, msgspec.DecodeError):
                # Skip failed data loads
                pass

        return data

    def _update_widgets(self, data: Dict[str, Any]) -> None:
        """Update all widgets with new data"""

        # Update dashboard
        dashboard = self.query_one(DashboardWidget)
        dashboard.update_data(data)

        # Update positions
        positions = self.query_one(PositionsWidget)
        positions.update_data(data.get("positions", {}))
        # Update balances
        balances = self.query_one(BalancesWidget)
        balances.update_data(data.get("balances", {}))


def run_monitor():
    """Entry point for the CLI monitor"""
    app = NexusTraderMonitor()
    app.run()


if __name__ == "__main__":
    run_monitor()



from typing import Any, Dict, List
from nexustrader.base import ExchangeManager
import ccxt
import msgspec
from nexustrader.config import BasicConfig
from nexustrader.exchange.okx.schema import OkxMarket
from nexustrader.exchange.okx.constants import OkxAccountType
from nexustrader.constants import AccountType, ConfigType
from nexustrader.schema import InstrumentId
from nexustrader.error import EngineBuildError


class OkxExchangeManager(ExchangeManager):
    api: ccxt.okx
    market: Dict[str, OkxMarket]  # symbol -> okx market
    market_id: Dict[str, str]  # symbol -> exchange symbol id

    def __init__(self, config: ConfigType | None = None):
        config = config or {}
        config["exchange_id"] = config.get("exchange_id", "okx")
        super().__init__(config)
        self.passphrase = config.get("password", None)
        self._public_conn_account_type = None

    def load_markets(self):
        market = self.api.load_markets()
        for symbol, mkt in market.items():
            try:
                mkt_json = msgspec.json.encode(mkt)
                mkt = msgspec.json.decode(mkt_json, type=OkxMarket)

                if (
                    mkt.spot or mkt.linear or mkt.inverse or mkt.future
                ) and not mkt.option:
                    symbol = self._parse_symbol(mkt, exchange_suffix="OKX")
                    mkt.symbol = symbol
                    self.market[symbol] = mkt
                    self.market_id[mkt.id] = (
                        symbol  # since okx symbol id is identical, no need to distinguish spot, linear, inverse
                    )

            except Exception as e:
                print(f"Error: {e}, {symbol}, {mkt}")
                continue

    def option(
        self,
        base: str | None = None,
        quote: str | None = None,
        exclude: List[str] | None = None,
    ) -> List[str]:
        raise NotImplementedError("Option is not supported for OKX")

    def validate_public_connector_config(
        self, account_type: AccountType, basic_config: BasicConfig | None = None
    ) -> None:
        """Validate public connector configuration for OKX exchange"""
        if not isinstance(account_type, OkxAccountType):
            raise EngineBuildError(f"Expected OkxAccountType, got {type(account_type)}")

        if basic_config.testnet != account_type.is_testnet:
            raise EngineBuildError(
                f"The `testnet` setting of OKX is not consistent with the public connector's account type `{account_type}`."
            )

    def validate_public_connector_limits(
        self, existing_connectors: Dict[AccountType, Any]
    ) -> None:
        """Validate public connector limits for OKX exchange"""
        okx_connectors = [
            c
            for c in existing_connectors.values()
            if hasattr(c, "account_type") and isinstance(c.account_type, OkxAccountType)
        ]
        if len(okx_connectors) > 1:
            raise EngineBuildError(
                "Only one public connector is supported for OKX, please remove the extra public connector config."
            )

    def set_public_connector_account_type(self, account_type: OkxAccountType) -> None:
        """Set the account type for public connector configuration"""
        self._public_conn_account_type = account_type

    def instrument_id_to_account_type(self, instrument_id: InstrumentId) -> AccountType:
        """Convert an instrument ID to the appropriate account type for OKX exchange"""
        if self._public_conn_account_type is None:
            raise EngineBuildError(
                "Public connector account type not set for OKX. Please add OKX in public_conn_config."
            )
        return self._public_conn_account_type



from typing import Optional
from nexustrader.schema import Order
from nexustrader.core.nautilius_core import Logger
from cachetools import TTLCache


class OrderRegistry:
    def __init__(
        self,
        ttl_maxsize: int = 72000,
        ttl_seconds: int = 3600,
    ):
        self._log = Logger(name=type(self).__name__)
        self._tmp_order: TTLCache[str, Order] = TTLCache(
            maxsize=ttl_maxsize, ttl=ttl_seconds
        )
        self._oids: set[str] = set()

    def register_order(self, oid: str) -> None:
        """Register an order to track its status"""
        self._log.debug(f"[ORDER REGISTER]: {oid}")
        self._oids.add(oid)

    def is_registered(self, oid: str) -> bool:
        """Check if an order is registered"""
        return oid in self._oids

    def unregister_order(self, oid: str) -> None:
        """Remove order mapping when no longer needed"""
        self._log.debug(f"[ORDER UNREGISTER]: {oid}")
        self._oids.discard(oid)

    def register_tmp_order(self, order: Order) -> None:
        """Register a temporary order"""
        self._tmp_order[order.oid] = order
        self._log.debug(f"[TMP ORDER REGISTER]: {order.oid}")

    def get_tmp_order(self, oid: str) -> Optional[Order]:
        self._log.debug(f"[TMP ORDER GET]: {oid}")
        return self._tmp_order.get(oid, None)



import os
import sys
from abc import abstractmethod
from typing import Literal, Dict, List, TypedDict, NotRequired
from enum import Enum
from dynaconf import Dynaconf
from throttled import Throttled as ThrottledSync
from throttled.asyncio import Throttled
from nautilus_trader.core.nautilus_pyo3 import LogColor  # noqa


def is_sphinx_build():
    return "sphinx" in sys.modules


if not os.path.exists(".keys/"):
    os.makedirs(".keys/")
if not os.path.exists(".keys/.secrets.toml") and not is_sphinx_build():
    raise FileNotFoundError(
        "Config file not found, please create a config file at .keys/.secrets.toml"
    )


settings = Dynaconf(
    envvar_prefix="NEXUS",
    settings_files=[".keys/settings.toml", ".keys/.secrets.toml"],
    load_dotenv=True,
)


def get_postgresql_config():
    """Get PostgreSQL configuration from environment variables or settings files."""
    return {
        "host": settings.get("PG_HOST", "localhost"),
        "port": settings.get("PG_PORT", 5432),
        "user": settings.get("PG_USER", "postgres"),
        "password": settings.get("PG_PASSWORD", ""),
        "database": settings.get("PG_DATABASE", "postgres"),
    }


def get_redis_config(in_docker: bool = False):
    try:
        if in_docker:
            return {
                "host": "redis",
                "db": settings.REDIS_DB,
                "password": settings.REDIS_PASSWORD,
            }

        return {
            "host": settings.REDIS_HOST,
            "port": settings.REDIS_PORT,
            "db": settings.REDIS_DB,
            "password": settings.REDIS_PASSWORD,
        }
    except Exception as e:
        raise ValueError(f"Failed to get Redis password: {e}")


IntervalType = Literal[
    "1s",
    "1m",
    "3m",
    "5m",
    "15m",
    "30m",
    "1h",
    "2h",
    "4h",
    "6h",
    "8h",
    "12h",
    "1d",
    "3d",
    "1w",
    "1M",
]


class BookLevel(Enum):
    L5 = "5"
    L10 = "10"
    L20 = "20"
    L50 = "50"


class KlineInterval(Enum):
    SECOND_1 = "1s"
    MINUTE_1 = "1m"
    MINUTE_3 = "3m"
    MINUTE_5 = "5m"
    MINUTE_15 = "15m"
    MINUTE_30 = "30m"
    HOUR_1 = "1h"
    HOUR_2 = "2h"
    HOUR_4 = "4h"
    HOUR_6 = "6h"
    HOUR_8 = "8h"
    HOUR_12 = "12h"
    DAY_1 = "1d"
    DAY_3 = "3d"
    WEEK_1 = "1w"
    MONTH_1 = "1M"

    @property
    def seconds(self) -> int:
        return {
            KlineInterval.SECOND_1: 1,
            KlineInterval.MINUTE_1: 60,
            KlineInterval.MINUTE_3: 180,
            KlineInterval.MINUTE_5: 300,
            KlineInterval.MINUTE_15: 900,
            KlineInterval.MINUTE_30: 1800,
            KlineInterval.HOUR_1: 3600,
            KlineInterval.HOUR_2: 7200,
            KlineInterval.HOUR_4: 14400,
            KlineInterval.HOUR_6: 21600,
            KlineInterval.HOUR_8: 28800,
            KlineInterval.HOUR_12: 43200,
            KlineInterval.DAY_1: 86400,
            KlineInterval.DAY_3: 259200,
            KlineInterval.WEEK_1: 604800,
            KlineInterval.MONTH_1: 2592000,
        }[self]

    @property
    def microseconds(self) -> int:
        return self.seconds * 1000


class SubmitType(Enum):
    CREATE = 0
    CANCEL = 1
    TWAP = 2
    CANCEL_TWAP = 3
    VWAP = 4
    CANCEL_VWAP = 5
    TAKE_PROFIT_AND_STOP_LOSS = 6
    MODIFY = 7
    CANCEL_ALL = 8
    BATCH = 9
    CREATE_WS = 10
    CANCEL_WS = 11


class EventType(Enum):
    BOOKL1 = 0
    TRADE = 1
    KLINE = 2
    MARK_PRICE = 3
    FUNDING_RATE = 4
    INDEX_PRICE = 5


class AlgoOrderStatus(Enum):
    RUNNING = "RUNNING"
    CANCELING = "CANCELING"
    FINISHED = "FINISHED"
    CANCELED = "CANCELED"
    FAILED = "FAILED"


class OrderStatus(Enum):
    # LOCAL
    INITIALIZED = "INITIALIZED"
    FAILED = "FAILED"
    CANCEL_FAILED = "CANCEL_FAILED"

    # IN-FLOW
    PENDING = "PENDING"
    CANCELING = "CANCELING"

    # OPEN
    ACCEPTED = "ACCEPTED"
    PARTIALLY_FILLED = "PARTIALLY_FILLED"

    # CLOSED
    FILLED = "FILLED"
    CANCELED = "CANCELED"
    EXPIRED = "EXPIRED"


class ExchangeType(Enum):
    BINANCE = "binance"
    OKX = "okx"
    BYBIT = "bybit"
    HYPERLIQUID = "hyperliquid"
    BITGET = "bitget"


class AccountType(Enum):
    pass


class OrderType(Enum):
    MARKET = "MARKET"
    LIMIT = "LIMIT"
    TAKE_PROFIT_MARKET = "TAKE_PROFIT_MARKET"
    TAKE_PROFIT_LIMIT = "TAKE_PROFIT_LIMIT"
    STOP_LOSS_MARKET = "STOP_LOSS_MARKET"
    STOP_LOSS_LIMIT = "STOP_LOSS_LIMIT"
    POST_ONLY = "POST_ONLY"

    @property
    def is_take_profit(self) -> bool:
        return self in (OrderType.TAKE_PROFIT_MARKET, OrderType.TAKE_PROFIT_LIMIT)

    @property
    def is_stop_loss(self) -> bool:
        return self in (OrderType.STOP_LOSS_MARKET, OrderType.STOP_LOSS_LIMIT)

    @property
    def is_market(self) -> bool:
        return self in (
            OrderType.MARKET,
            OrderType.TAKE_PROFIT_MARKET,
            OrderType.STOP_LOSS_MARKET,
        )

    @property
    def is_limit(self) -> bool:
        return self in (
            OrderType.LIMIT,
            OrderType.TAKE_PROFIT_LIMIT,
            OrderType.STOP_LOSS_LIMIT,
        )

    @property
    def is_post_only(self) -> bool:
        return self == OrderType.POST_ONLY


class TriggerType(Enum):
    LAST_PRICE = "LAST_PRICE"
    MARK_PRICE = "MARK_PRICE"
    INDEX_PRICE = "INDEX_PRICE"


class OrderSide(Enum):
    BUY = "BUY"
    SELL = "SELL"

    @property
    def is_buy(self) -> bool:
        return self == OrderSide.BUY

    @property
    def is_sell(self) -> bool:
        return self == OrderSide.SELL


class TimeInForce(Enum):
    GTC = "GTC"
    IOC = "IOC"
    FOK = "FOK"


class PositionSide(Enum):
    LONG = "LONG"
    SHORT = "SHORT"
    FLAT = "FLAT"

    @property
    def is_long(self) -> bool:
        return self == PositionSide.LONG

    @property
    def is_short(self) -> bool:
        return self == PositionSide.SHORT

    @property
    def is_flat(self) -> bool:
        return self == PositionSide.FLAT


class InstrumentType(Enum):
    SPOT = "spot"
    MARGIN = "margin"
    FUTURE = "future"
    OPTION = "option"
    SWAP = "swap"
    LINEAR = "linear"
    INVERSE = "inverse"


class OptionType(Enum):
    CALL = "call"
    PUT = "put"


STATUS_TRANSITIONS: Dict[OrderStatus, List[OrderStatus]] = {
    OrderStatus.PENDING: [
        OrderStatus.PENDING,
        OrderStatus.CANCELED,
        OrderStatus.CANCELING,
        OrderStatus.ACCEPTED,
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
        OrderStatus.CANCEL_FAILED,
    ],
    OrderStatus.CANCELING: [
        OrderStatus.CANCELED,
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
    ],
    OrderStatus.ACCEPTED: [
        OrderStatus.ACCEPTED,  # for modify order, pending -> accepted -> accepted -> (not allowed) -> pending
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
        OrderStatus.CANCELING,
        OrderStatus.CANCELED,
        OrderStatus.EXPIRED,
        OrderStatus.CANCEL_FAILED,
    ],
    OrderStatus.PARTIALLY_FILLED: [
        OrderStatus.ACCEPTED,  # for modify order, accepted -> partially filled -> accepted -> (not allowed) -> pending
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
        OrderStatus.CANCELING,
        OrderStatus.CANCELED,
        OrderStatus.EXPIRED,
        OrderStatus.CANCEL_FAILED,
    ],
    OrderStatus.CANCEL_FAILED: [OrderStatus.FILLED],
    OrderStatus.FILLED: [],
    OrderStatus.CANCELED: [],
    OrderStatus.EXPIRED: [],
    OrderStatus.FAILED: [],
}


class DataType(Enum):
    BOOKL1 = "bookl1"
    BOOKL2 = "bookl2"
    TRADE = "trade"
    KLINE = "kline"
    MARK_PRICE = "mark_price"
    FUNDING_RATE = "funding_rate"
    INDEX_PRICE = "index_price"


class StorageType(Enum):
    SQLITE = "sqlite"
    POSTGRESQL = "postgresql"


class RateLimiter:
    @abstractmethod
    def __call__(self, endpoint: str) -> Throttled: ...


class RateLimiterSync:
    @abstractmethod
    def __call__(self, endpoint: str) -> ThrottledSync: ...


class ConfigType(TypedDict):
    apiKey: str
    secret: str
    exchange_id: str
    sandbox: bool
    password: NotRequired[str]



import msgspec
from typing import Dict, Any
import base64
import httpx
from urllib.parse import urlencode
from nexustrader.base import ApiClient, RetryManager
from nexustrader.exchange.okx.constants import (
    OkxRateLimiter,
    OkxRateLimiterSync,
)
from nexustrader.exchange.okx.error import OkxHttpError, OkxRequestError, retry_check
from nexustrader.exchange.okx.schema import (
    OkxPlaceOrderResponse,
    OkxCancelOrderResponse,
    OkxGeneralResponse,
    OkxErrorResponse,
    OkxBalanceResponse,
    OkxPositionResponse,
    OkxCandlesticksResponse,
    OkxSavingsBalanceResponse,
    OkxSavingsPurchaseRedemptResponse,
    OkxSavingsLendingRateSummaryResponse,
    OkxSavingsLendingRateHistoryResponse,
    OkxAssetTransferResponse,
    OkxAmendOrderResponse,
    OkxFinanceStakingDefiRedeemResponse,
    OkxFinanceStakingDefiPurchaseResponse,
    OkxFinanceStakingDefiOffersResponse,
    OkxAccountConfigResponse,
    OkxIndexCandlesticksResponse,
    OkxBatchOrderResponse,
    OkxCancelBatchOrderResponse,
    OkxTickersResponse,
    OkxOrderResponse,
)
from nexustrader.core.nautilius_core import hmac_signature, LiveClock


class OkxApiClient(ApiClient):
    _limiter: OkxRateLimiter
    _limiter_sync: OkxRateLimiterSync

    def __init__(
        self,
        clock: LiveClock,
        api_key: str = None,
        secret: str = None,
        passphrase: str = None,
        testnet: bool = False,
        timeout: int = 10,
        enable_rate_limit: bool = True,
        max_retries: int = 0,
        delay_initial_ms: int = 100,
        delay_max_ms: int = 800,
        backoff_factor: int = 2,
    ):
        super().__init__(
            clock=clock,
            api_key=api_key,
            secret=secret,
            timeout=timeout,
            rate_limiter=OkxRateLimiter(enable_rate_limit),
            rate_limiter_sync=OkxRateLimiterSync(enable_rate_limit),
            retry_manager=RetryManager(
                max_retries=max_retries,
                delay_initial_ms=delay_initial_ms,
                delay_max_ms=delay_max_ms,
                backoff_factor=backoff_factor,
                exc_types=(OkxRequestError, httpx.NetworkError),
                retry_check=retry_check,
            ),
        )

        self._base_url = "https://www.okx.com"
        self._passphrase = passphrase
        self._testnet = testnet
        self._place_order_decoder = msgspec.json.Decoder(OkxPlaceOrderResponse)
        self._cancel_order_decoder = msgspec.json.Decoder(OkxCancelOrderResponse)
        self._general_response_decoder = msgspec.json.Decoder(OkxGeneralResponse)
        self._error_response_decoder = msgspec.json.Decoder(OkxErrorResponse)
        self._balance_response_decoder = msgspec.json.Decoder(
            OkxBalanceResponse, strict=False
        )
        self._position_response_decoder = msgspec.json.Decoder(
            OkxPositionResponse, strict=False
        )
        self._candles_response_decoder = msgspec.json.Decoder(
            OkxCandlesticksResponse, strict=False
        )
        self._index_candles_response_decoder = msgspec.json.Decoder(
            OkxIndexCandlesticksResponse, strict=False
        )
        self._savings_balance_response_decoder = msgspec.json.Decoder(
            OkxSavingsBalanceResponse, strict=False
        )

        self._savings_purchase_redempt_response_decoder = msgspec.json.Decoder(
            OkxSavingsPurchaseRedemptResponse, strict=False
        )

        self._savings_lending_rate_summary_response_decoder = msgspec.json.Decoder(
            OkxSavingsLendingRateSummaryResponse, strict=False
        )

        self._savings_lending_rate_history_response_decoder = msgspec.json.Decoder(
            OkxSavingsLendingRateHistoryResponse, strict=False
        )

        self._asset_transfer_response_decoder = msgspec.json.Decoder(
            OkxAssetTransferResponse, strict=False
        )

        self._amend_order_response_decoder = msgspec.json.Decoder(
            OkxAmendOrderResponse, strict=False
        )

        self._finance_staking_defi_redeem_response_decoder = msgspec.json.Decoder(
            OkxFinanceStakingDefiRedeemResponse, strict=False
        )

        self._finance_staking_defi_purchase_response_decoder = msgspec.json.Decoder(
            OkxFinanceStakingDefiPurchaseResponse, strict=False
        )

        self._finance_staking_defi_offers_response_decoder = msgspec.json.Decoder(
            OkxFinanceStakingDefiOffersResponse, strict=False
        )

        self._account_config_response_decoder = msgspec.json.Decoder(
            OkxAccountConfigResponse, strict=False
        )

        self._batch_order_response_decoder = msgspec.json.Decoder(OkxBatchOrderResponse)

        self._cancel_batch_order_response_decoder = msgspec.json.Decoder(
            OkxCancelBatchOrderResponse
        )

        self._tickers_response_decoder = msgspec.json.Decoder(OkxTickersResponse)

        self._order_response_decoder = msgspec.json.Decoder(
            OkxOrderResponse, strict=False
        )

        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": "TradingBot/1.0",
        }

        if self._testnet:
            self._headers["x-simulated-trading"] = "1"

    def get_api_v5_account_balance(self, ccy: str | None = None) -> OkxBalanceResponse:
        """
        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-balance
        """
        endpoint = "/api/v5/account/balance"
        payload = {"ccy": ccy} if ccy else {}
        cost = self._get_rate_limit_cost(1)
        self._limiter_sync(endpoint).limit(key=endpoint, cost=cost)
        raw = self._fetch_sync("GET", endpoint, payload=payload, signed=True)
        return self._balance_response_decoder.decode(raw)

    def get_api_v5_account_positions(
        self,
        inst_type: str | None = None,
        inst_id: str | None = None,
        pos_id: str | None = None,
    ) -> OkxPositionResponse:
        """
        https://www.okx.com/docs-v5/en/#trading-account-rest-api-get-positions
        """
        endpoint = "/api/v5/account/positions"
        payload = {
            k: v
            for k, v in {
                "instType": inst_type,
                "instId": inst_id,
                "posId": pos_id,
            }.items()
            if v is not None
        }
        cost = self._get_rate_limit_cost(1)
        self._limiter_sync(endpoint).limit(key=endpoint, cost=cost)
        raw = self._fetch_sync("GET", endpoint, payload=payload, signed=True)
        return self._position_response_decoder.decode(raw)

    async def post_api_v5_trade_order(
        self,
        inst_id: str,
        td_mode: str,
        side: str,
        ord_type: str,
        sz: str,
        **kwargs,
    ) -> OkxPlaceOrderResponse:
        """
        Place a new order
        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-place-order

        {'arg': {'channel': 'orders', 'instType': 'ANY', 'uid': '611800569950521616'}, 'data': [{'instType': 'SWAP', 'instId': 'BTC-USDT-SWAP', 'tgtCcy': '', 'ccy': '', 'ordId': '1993784914940116992', 'clOrdId': '', 'algoClOrdId': '', 'algoId': '', 'tag': '', 'px': '80000', 'sz': '0.1', 'notionalUsd': '80.0128', 'ordType': 'limit', 'side': 'buy', 'posSide': 'long', 'tdMode': 'cross', 'accFillSz': '0', 'fillNotionalUsd': '', 'avgPx': '0', 'state': 'canceled', 'lever': '3', 'pnl': '0', 'feeCcy': 'USDT', 'fee': '0', 'rebateCcy': 'USDT', 'rebate': '0', 'category': 'normal', 'uTime': '1731921825881', 'cTime': '1731921820806', 'source': '', 'reduceOnly': 'false', 'cancelSource': '1', 'quickMgnType': '', 'stpId': '', 'stpMode': 'cancel_maker', 'attachAlgoClOrdId': '', 'lastPx': '91880', 'isTpLimit': 'false', 'slTriggerPx': '', 'slTriggerPxType': '', 'tpOrdPx': '', 'tpTriggerPx': '', 'tpTriggerPxType': '', 'slOrdPx': '', 'fillPx': '', 'tradeId': '', 'fillSz': '0', 'fillTime': '', 'fillPnl': '0', 'fillFee': '0', 'fillFeeCcy': '', 'execType': '', 'fillPxVol': '', 'fillPxUsd': '', 'fillMarkVol': '', 'fillFwdPx': '', 'fillMarkPx': '', 'amendSource': '', 'reqId': '', 'amendResult': '', 'code': '0', 'msg': '', 'pxType': '', 'pxUsd': '', 'pxVol': '', 'linkedAlgoOrd': {'algoId': ''}, 'attachAlgoOrds': []}]}
        """
        endpoint = "/api/v5/trade/order"
        payload = {
            "instId": inst_id,
            "tdMode": td_mode,
            "side": side,
            "ordType": ord_type,
            "sz": sz,
            **kwargs,
        }
        cost = self._get_rate_limit_cost(1)
        await self._limiter(endpoint).limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._place_order_decoder.decode(raw)

    async def post_api_v5_trade_cancel_order(
        self, instId: str, clOrdId: str
    ) -> OkxCancelOrderResponse:
        """
        Cancel an existing order
        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-cancel-order
        """
        endpoint = "/api/v5/trade/cancel-order"
        payload = {"instId": instId, "clOrdId": clOrdId}

        cost = self._get_rate_limit_cost(1)
        await self._limiter(endpoint).limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._cancel_order_decoder.decode(raw)

    async def post_api_v5_trade_cancel_batch_order(
        self, payload: list[dict]
    ) -> OkxCancelBatchOrderResponse:
        """
        Cancel multiple orders in batch
        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-cancel-batch-orders

        Args:
            payload: List of dictionaries, each containing:
                - instId (str): Instrument ID, e.g. BTC-USDT
                - ordId (str, optional): Order ID
                - clOrdId (str, optional): Client Order ID
                Note: Either ordId or clOrdId is required for each order

        Returns:
            OkxCancelBatchOrderResponse: Response containing cancellation results
        """
        endpoint = "/api/v5/trade/cancel-batch-orders"
        cost = len(payload)
        await self._limiter(endpoint).limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._cancel_batch_order_response_decoder.decode(raw)

    def get_api_v5_market_history_index_candles(
        self,
        instId: str,
        bar: str | None = None,
        after: str | None = None,
        before: str | None = None,
        limit: int | None = None,
    ) -> OkxIndexCandlesticksResponse:
        """
        GET /api/v5/market/history-index-candles
        """
        endpoint = "/api/v5/market/history-index-candles"
        payload = {
            "instId": instId,
            "bar": bar.replace("candle", ""),
            "after": after,
            "before": before,
            "limit": str(limit),
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        cost = self._get_rate_limit_cost(1)
        self._limiter_sync(endpoint).limit(key=endpoint, cost=cost)
        raw = self._fetch_sync("GET", endpoint, payload=payload, signed=False)
        return self._index_candles_response_decoder.decode(raw)

    def get_api_v5_market_candles(
        self,
        instId: str,
        bar: str | None = None,
        after: str | None = None,
        before: str | None = None,
        limit: int | None = None,
    ) -> OkxCandlesticksResponse:
        # the default bar is 1m
        endpoint = "/api/v5/market/candles"
        payload = {
            k: v
            for k, v in {
                "instId": instId,
                "bar": bar.replace("candle", "") if bar else None,
                "after": after,
                "before": before,
                "limit": str(limit),
            }.items()
            if v is not None
        }
        cost = self._get_rate_limit_cost(1)
        self._limiter_sync(endpoint).limit(key=endpoint, cost=cost)
        raw = self._fetch_sync("GET", endpoint, payload=payload, signed=False)
        return self._candles_response_decoder.decode(raw)

    def get_api_v5_market_history_candles(
        self,
        instId: str,
        bar: str | None = None,
        after: str | None = None,
        before: str | None = None,
        limit: str | None = None,
    ) -> OkxCandlesticksResponse:
        # the default bar is 1m
        endpoint = "/api/v5/market/history-candles"
        payload = {
            k: v
            for k, v in {
                "instId": instId,
                "bar": bar.replace("candle", ""),
                "after": after,
                "before": before,
                "limit": str(limit),
            }.items()
            if v is not None
        }
        cost = self._get_rate_limit_cost(1)
        self._limiter_sync(endpoint).limit(key=endpoint, cost=cost)
        raw = self._fetch_sync("GET", endpoint, payload=payload, signed=False)
        return self._candles_response_decoder.decode(raw)

    async def get_api_v5_finance_savings_balance(
        self,
        ccy: str | None = None,
    ) -> OkxSavingsBalanceResponse:
        """
        GET /api/v5/finance/savings/balance
        """
        endpoint = "/api/v5/finance/savings/balance"
        payload = {"ccy": ccy} if ccy else None
        raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
        return self._savings_balance_response_decoder.decode(raw)

    async def post_api_v5_finance_savings_purchase_redempt(
        self,
        ccy: str,
        amt: str,
        side: str,
        rate: str | None = None,
    ) -> OkxSavingsPurchaseRedemptResponse:
        """
        POST /api/v5/finance/savings/purchase-redempt
        """
        endpoint = "/api/v5/finance/savings/purchase-redempt"
        payload = {
            "ccy": ccy,
            "amt": amt,
            "side": side,
        }
        if rate:
            payload["rate"] = rate

        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._savings_purchase_redempt_response_decoder.decode(raw)

    async def get_api_v5_finance_savings_lending_rate_summary(
        self,
        ccy: str | None = None,
    ) -> OkxSavingsLendingRateSummaryResponse:
        """
        /api/v5/finance/savings/lending-rate-summary
        """
        endpoint = "/api/v5/finance/savings/lending-rate-summary"
        payload = {"ccy": ccy} if ccy else None
        raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
        return self._savings_lending_rate_summary_response_decoder.decode(raw)

    async def get_api_v5_finance_savings_lending_rate_history(
        self,
        ccy: str | None = None,
        after: str | None = None,
        before: str | None = None,
        limit: str | None = None,
    ) -> OkxSavingsLendingRateHistoryResponse:
        """
        GET /api/v5/finance/savings/lending-rate-history
        """
        endpoint = "/api/v5/finance/savings/lending-rate-history"
        payload = {
            "ccy": ccy,
            "after": after,
            "before": before,
            "limit": limit,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
        return self._savings_lending_rate_history_response_decoder.decode(raw)

    async def post_api_v5_asset_transfer(
        self,
        ccy: str,
        amt: str,
        from_acct: str,  # from
        to_acct: str,  # to
        type: str = "0",
        subAcct: str = None,
        loanTrans: bool = False,
        omitPosRisk: bool = False,
        clientId: str = None,
    ) -> OkxAssetTransferResponse:
        """
        POST /api/v5/asset/transfer
        """
        endpoint = "/api/v5/asset/transfer"
        payload = {
            "ccy": ccy,
            "amt": amt,
            "from": from_acct,
            "to": to_acct,
            "type": type,
            "subAcct": subAcct,
            "loanTrans": loanTrans,
            "omitPosRisk": omitPosRisk,
            "clientId": clientId,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._asset_transfer_response_decoder.decode(raw)

    async def post_api_v5_trade_amend_order(
        self,
        instId: str,
        cxlOnFail: bool = False,
        ordId: str = None,
        clOrdId: str = None,
        reqId: str = None,
        newSz: str = None,
        newPx: str = None,
        newPxUsd: str = None,
        newPxVol: str = None,
        attachAlgoOrds: list[dict] = None,
    ):
        """
        https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-amend-order
        POST /api/v5/trade/amend-order
        """
        endpoint = "/api/v5/trade/amend-order"
        payload = {
            "instId": instId,
            "cxlOnFail": cxlOnFail,
            "ordId": ordId,
            "clOrdId": clOrdId,
            "reqId": reqId,
            "newSz": newSz,
            "newPx": newPx,
            "newPxUsd": newPxUsd,
            "newPxVol": newPxVol,
            "attachAlgoOrds": attachAlgoOrds,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        cost = self._get_rate_limit_cost(1)
        await self._limiter(endpoint).limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._amend_order_response_decoder.decode(raw)

    async def post_api_v5_finance_staking_defi_redeem(
        self, ordId: str, protocolType: str, allowEarlyRedeem: bool = False
    ):
        """
        POST /api/v5/finance/staking-defi/redeem
        """
        endpoint = "/api/v5/finance/staking-defi/redeem"
        payload = {
            "ordId": ordId,
            "protocolType": protocolType,
            "allowEarlyRedeem": allowEarlyRedeem,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._finance_staking_defi_redeem_response_decoder.decode(raw)

    async def post_api_v5_trade_batch_orders(
        self, payload: list[dict]
    ) -> OkxBatchOrderResponse:
        """
        POST /api/v5/trade/batch-orders
        """
        endpoint = "/api/v5/trade/batch-orders"
        cost = len(payload)
        await self._limiter(endpoint).limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._batch_order_response_decoder.decode(raw)

    async def post_api_v5_finance_staking_defi_purchase(
        self, productId: str, investData: list[dict], term: str = None, tag: str = None
    ):
        """
        productId	String	Yes	Product ID
        investData	Array of objects	Yes	Investment data
        > ccy	String	Yes	Investment currency, e.g. BTC
        > amt	String	Yes	Investment amount
        term	String	Conditional	Investment term
        Investment term must be specified for fixed-term product
        tag	String	No	Order tag
        A combination of case-sensitive alphanumerics, all numbers, or all letters of up to 16 characters.
        """

        endpoint = "/api/v5/finance/staking-defi/purchase"
        payload = {
            "productId": productId,
            "investData": investData,
            "term": term,
            "tag": tag,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
        return self._finance_staking_defi_purchase_response_decoder.decode(raw)

    async def get_api_v5_finance_staking_defi_offers(
        self, productId: str = None, protocolType: str = None, ccy: str = None
    ):
        endpoint = "/api/v5/finance/staking-defi/offers"
        payload = {
            "productId": productId,
            "protocolType": protocolType,
            "ccy": ccy,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
        return self._finance_staking_defi_offers_response_decoder.decode(raw)

    def get_api_v5_account_config(self):
        """
        GET /api/v5/account/config
        """
        endpoint = "/api/v5/account/config"
        raw = self._fetch_sync("GET", endpoint, signed=True)
        cost = self._get_rate_limit_cost(1)
        self._limiter_sync(endpoint).limit(key=endpoint, cost=cost)
        return self._account_config_response_decoder.decode(raw)

    def _generate_signature(self, message: str) -> str:
        hex_digest = hmac_signature(self._secret, message)
        digest = bytes.fromhex(hex_digest)
        return base64.b64encode(digest).decode()

    def _get_signature(
        self, ts: str, method: str, request_path: str, payload: bytes
    ) -> str:
        body = payload.decode() if payload else ""
        sign_str = f"{ts}{method}{request_path}{body}"
        signature = self._generate_signature(sign_str)
        return signature

    def _get_timestamp(self) -> str:
        return (
            self._clock.utc_now()
            .isoformat(timespec="milliseconds")
            .replace("+00:00", "Z")
        )

    def _get_headers(
        self, ts: str, method: str, request_path: str, payload: bytes
    ) -> Dict[str, Any]:
        headers = self._headers
        signature = self._get_signature(ts, method, request_path, payload)
        headers.update(
            {
                "OK-ACCESS-KEY": self._api_key,
                "OK-ACCESS-SIGN": signature,
                "OK-ACCESS-TIMESTAMP": ts,
                "OK-ACCESS-PASSPHRASE": self._passphrase,
            }
        )
        return headers

    async def _fetch(
        self,
        method: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ) -> bytes:
        """
        Fetch data from the OKX API asynchronously.
        """
        return await self._retry_manager.run(
            name=f"{method} {endpoint}",
            func=self._fetch_async,
            method=method,
            endpoint=endpoint,
            payload=payload,
            signed=signed,
        )

    async def _fetch_async(
        self,
        method: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ) -> bytes:
        self._init_session(self._base_url)

        request_path = endpoint
        headers = self._headers
        timestamp = self._get_timestamp()

        payload = payload or {}

        payload_json = (
            urlencode(payload) if method == "GET" else msgspec.json.encode(payload)
        )

        if method == "GET":
            if payload_json:
                request_path += f"?{payload_json}"
            payload_json = None

        if signed and self._api_key:
            headers = self._get_headers(timestamp, method, request_path, payload_json)

        try:
            self._log.debug(
                f"{method} {request_path} Headers: {headers} payload: {payload_json}"
            )

            response = await self._session.request(
                method=method,
                url=request_path,
                headers=headers,
                data=payload_json,
            )
            raw = response.content

            if response.status_code >= 400:
                raise OkxHttpError(
                    status_code=response.status_code,
                    message=msgspec.json.decode(raw),
                    headers=response.headers,
                )
            okx_response = self._general_response_decoder.decode(raw)
            if okx_response.code == "0":
                return raw
            else:
                okx_error_response = self._error_response_decoder.decode(raw)
                for data in okx_error_response.data:
                    raise OkxRequestError(
                        error_code=data.sCode,
                        status_code=response.status_code,
                        message=data.sMsg,
                    )
                raise OkxRequestError(
                    error_code=okx_error_response.code,
                    status_code=response.status_code,
                    message=okx_error_response.msg,
                )
        except httpx.TimeoutException as e:
            self._log.error(f"Timeout {method} {request_path} {e}")
            raise
        except httpx.ConnectError as e:
            self._log.error(f"Connection Error {method} {request_path} {e}")
            raise
        except httpx.HTTPStatusError as e:
            self._log.error(f"HTTP Error {method} {request_path} {e}")
            raise
        except httpx.RequestError as e:
            self._log.error(f"Request Error {method} {request_path} {e}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} {request_path} {e}")
            raise

    def _fetch_sync(
        self,
        method: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ) -> bytes:
        self._init_sync_session(self._base_url)

        request_path = endpoint
        headers = self._headers
        timestamp = self._get_timestamp()

        payload = payload or {}

        payload_json = (
            urlencode(payload) if method == "GET" else msgspec.json.encode(payload)
        )

        if method == "GET":
            if payload_json:
                request_path += f"?{payload_json}"
            payload_json = None

        if signed and self._api_key:
            headers = self._get_headers(timestamp, method, request_path, payload_json)

        try:
            self._log.debug(
                f"{method} {request_path} Headers: {headers} payload: {payload_json}"
            )

            response = self._sync_session.request(
                method=method,
                url=request_path,
                headers=headers,
                data=payload_json,
            )
            raw = response.content

            if response.status_code >= 400:
                raise OkxHttpError(
                    status_code=response.status_code,
                    message=msgspec.json.decode(raw),
                    headers=response.headers,
                )
            okx_response = self._general_response_decoder.decode(raw)
            if okx_response.code == "0":
                return raw
            else:
                okx_error_response = self._error_response_decoder.decode(raw)
                for data in okx_error_response.data:
                    raise OkxRequestError(
                        error_code=data.sCode,
                        status_code=response.status_code,
                        message=data.sMsg,
                    )
                raise OkxRequestError(
                    error_code=okx_error_response.code,
                    status_code=response.status_code,
                    message=okx_error_response.msg,
                )
        except httpx.TimeoutException as e:
            self._log.error(f"Timeout {method} {request_path} {e}")
            raise
        except httpx.ConnectError as e:
            self._log.error(f"Connection Error {method} {request_path} {e}")
            raise
        except httpx.HTTPStatusError as e:
            self._log.error(f"HTTP Error {method} {request_path} {e}")
            raise
        except httpx.RequestError as e:
            self._log.error(f"Request Error {method} {request_path} {e}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} {request_path} {e}")
            raise

    def get_api_v5_market_tickers(
        self,
        inst_type: str,
        uly: str | None = None,
        inst_family: str | None = None,
    ) -> OkxTickersResponse:
        """
        GET /api/v5/market/tickers

        Retrieve the latest price snapshot, best bid/ask price, and trading volume
        in the last 24 hours for multiple instruments.

        Rate Limit: 20 requests per 2 seconds

        Args:
            inst_type: Instrument type (SPOT, SWAP, FUTURES, OPTION)
            uly: Underlying (optional), e.g. BTC-USD. Applicable to FUTURES/SWAP/OPTION
            inst_family: Instrument family (optional). Applicable to FUTURES/SWAP/OPTION

        Returns:
            OkxTickersResponse: Response containing ticker data for multiple instruments
        """
        endpoint = "/api/v5/market/tickers"
        payload = {
            "instType": inst_type,
            "uly": uly,
            "instFamily": inst_family,
        }
        payload = {k: v for k, v in payload.items() if v is not None}

        cost = self._get_rate_limit_cost(1)
        self._limiter_sync(endpoint).limit(key=endpoint, cost=cost)
        raw = self._fetch_sync("GET", endpoint, payload=payload, signed=False)
        return self._tickers_response_decoder.decode(raw)

    def get_api_v5_market_ticker(self, inst_id: str) -> OkxTickersResponse:
        """
        GET /api/v5/market/ticker

        Retrieve the latest price snapshot, best bid/ask price, and trading volume
        in the last 24 hours for a single instrument.

        Rate Limit: 20 requests per 2 seconds

        Args:
            inst_id: Instrument ID (e.g., BTC-USD-SWAP)

        Returns:
            OkxTickersResponse: Response containing ticker data for the single instrument
        """
        endpoint = "/api/v5/market/ticker"
        payload = {"instId": inst_id}

        cost = self._get_rate_limit_cost(1)
        self._limiter_sync(endpoint).limit(key=endpoint, cost=cost)
        raw = self._fetch_sync("GET", endpoint, payload=payload, signed=False)
        return self._tickers_response_decoder.decode(raw)

    async def get_api_v5_trade_order(
        self,
        inst_id: str,
        ord_id: str | None = None,
        cl_ord_id: str | None = None,
    ) -> OkxOrderResponse:
        """
        GET /api/v5/trade/order

        Retrieve order details. Either ordId or clOrdId is required.
        If both are passed, ordId will be used.

        Rate Limit: 60 requests per 2 seconds

        Args:
            inst_id: Instrument ID (e.g., BTC-USDT). Only applicable to live instruments
            ord_id: Order ID. Either ordId or clOrdId is required
            cl_ord_id: Client Order ID as assigned by the client. Either ordId or clOrdId is required

        Returns:
            OkxOrderResponse: Response containing order details

        Raises:
            ValueError: If neither ord_id nor cl_ord_id is provided
        """
        if not ord_id and not cl_ord_id:
            raise ValueError("Either ord_id or cl_ord_id must be provided")

        endpoint = "/api/v5/trade/order"
        payload = {"instId": inst_id}

        if ord_id:
            payload["ordId"] = ord_id
        if cl_ord_id:
            payload["clOrdId"] = cl_ord_id

        cost = self._get_rate_limit_cost(1)
        await self._limiter(endpoint).limit(key=endpoint, cost=cost)
        raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
        return self._order_response_decoder.decode(raw)



import msgspec
from decimal import Decimal
from nexustrader.schema import BaseMarket, Balance
from nexustrader.exchange.hyperliquid.constants import (
    HyperLiquidKlineInterval,
    HyperLiquidOrderStatusType,
    HyperLiquidOrderSide,
    HyperLiquidFillDirection,
)


class HyperLiquidMarketInfo(msgspec.Struct, kw_only=True):
    """Market information from HyperLiquid exchange
    {



            "circulatingSupply": "8888888887.9898376465",
            "coin": "@153",
            "totalSupply": "8888888887.9898376465",
            "tokens": [
                241,
                0
            ],

            "index": 153,
            "isCanonical": false
        },

     "info": {
            "szDecimals": 5,

            "maxLeverage": 40,
            "funding": "0.0000069988",
            "openInterest": "9584.3844",


            "premium": "-0.0004322507",
            "oraclePx": "83285.0",

            "impactPxs": [
                "83238.0",
                "83249.0"
            ],
            "baseId": 0
        },


    """

    # Common fields
    name: str
    prevDayPx: str
    dayNtlVlm: str
    markPx: str
    midPx: str | None = None
    dayBaseVlm: str

    # Spot specific fields
    circulatingSupply: str | None = None
    coin: str | None = None
    totalSupply: str | None = None
    tokens: list[str] | None = None
    index: str | None = None
    isCanonical: bool | None = None

    # Perpetual specific fields
    szDecimals: str | None = None
    maxLeverage: str | None = None
    funding: str | None = None
    openInterest: str | None = None
    premium: str | None = None
    oraclePx: str | None = None
    impactPxs: list[str] | None = None
    baseId: int | None = None


class HyperLiquidMarket(BaseMarket):
    info: HyperLiquidMarketInfo
    baseName: str


class HyperLiquidOrderRestingStatus(msgspec.Struct):
    oid: int  # Order ID
    cloid: str | None = None  # Client order ID


class HyperLiquidOrderFilledStatus(msgspec.Struct):
    oid: int  # Order ID
    avgPx: str
    totalSz: str
    cloid: str | None = None  # Client order ID


class HyperLiquidOrderStatus(msgspec.Struct, kw_only=True, omit_defaults=True):
    """Order status information"""

    error: str | None = None  # Error message if any
    resting: HyperLiquidOrderRestingStatus | None = (
        None  # Contains oid when order is resting
    )
    filled: HyperLiquidOrderFilledStatus | None = (
        None  # Contains oid when order is filled
    )


class HyperLiquidOrderData(msgspec.Struct):
    """Order response data"""

    statuses: list[HyperLiquidOrderStatus]


class HyperLiquidOrderResponseData(msgspec.Struct):
    """Order response wrapper"""

    type: str
    data: HyperLiquidOrderData


class HyperLiquidOrderResponse(msgspec.Struct):
    """Response from order placement"""

    status: str
    response: HyperLiquidOrderResponseData


class HyperLiquidOrderStatusCancelStatus(msgspec.Struct):
    error: str


class HyperLiquidOrderDataStatuses(msgspec.Struct):
    statuses: list[str | HyperLiquidOrderStatusCancelStatus]


class HyperLiquidCancelResponseData(msgspec.Struct):
    type: str
    data: HyperLiquidOrderDataStatuses


class HyperLiquidCancelResponse(msgspec.Struct):
    """Response from order cancellation"""

    status: str
    response: HyperLiquidCancelResponseData


class HyperLiquidCumFunding(msgspec.Struct):
    """Cumulative funding information"""

    allTime: str
    sinceChange: str
    sinceOpen: str


class HyperLiquidLeverage(msgspec.Struct):
    """Leverage information"""

    type: str
    value: int
    rawUsd: str | None = None


class HyperLiquidActiveAssetData(msgspec.Struct):
    """Active asset data for a specific coin"""

    user: str  # User address
    coin: str  # Coin/token symbol
    leverage: HyperLiquidLeverage  # Leverage information
    maxTradeSzs: list[str]  # Maximum trade sizes
    availableToTrade: list[str]  # Available amounts to trade
    markPx: str  # Mark price


class HyperLiquidPosition(msgspec.Struct, kw_only=True):
    """Individual position data"""

    coin: str
    cumFunding: HyperLiquidCumFunding
    entryPx: str
    leverage: HyperLiquidLeverage
    liquidationPx: str | None = None
    marginUsed: str
    maxLeverage: int
    positionValue: str
    returnOnEquity: str
    szi: str
    unrealizedPnl: str


class HyperLiquidAssetPosition(msgspec.Struct):
    """Asset position wrapper"""

    position: HyperLiquidPosition
    type: str


class HyperLiquidMarginSummary(msgspec.Struct):
    """Margin summary information"""

    accountValue: str
    totalMarginUsed: str
    totalNtlPos: str
    totalRawUsd: str


class HyperLiquidUserPerpsSummary(msgspec.Struct):
    """User perpetuals summary information"""

    assetPositions: list[HyperLiquidAssetPosition]
    crossMaintenanceMarginUsed: str
    crossMarginSummary: HyperLiquidMarginSummary
    marginSummary: HyperLiquidMarginSummary
    time: int
    withdrawable: str


class HyperLiquidMeta(msgspec.Struct):
    """Market metadata"""

    universe: list[dict]
    amms: list[dict]


class HyperLiquidKline(msgspec.Struct):
    """
    Kline/candlestick data from HyperLiquid API
    [
        {
            "T": 1681924499999,
            "c": "29258.0",
            "h": "29309.0",
            "i": "15m",
            "l": "29250.0",
            "n": 189,
            "o": "29295.0",
            "s": "BTC",
            "t": 1681923600000,
            "v": "0.98639"
        }
    ]
    """

    T: int  # Close time
    c: str  # Close price
    h: str  # High price
    i: str  # Interval
    l: str  # Low price # noqa: E741
    n: int  # Number of trades
    o: str  # Open price
    s: str  # Symbol
    t: int  # Open time
    v: str  # Volume


class HyperLiquidTrade(msgspec.Struct):
    """Trade data"""

    timestamp: int
    price: str
    size: str
    side: str
    hash: str


class HyperLiquidOrderBook(msgspec.Struct):
    """Order book data"""

    levels: list[list[str]]  # [price, size] pairs
    timestamp: int


class HyperLiquidTicker(msgspec.Struct):
    """Ticker data"""

    symbol: str
    lastPrice: str
    volume24h: str
    priceChange24h: str
    high24h: str
    low24h: str


class HyperLiquidUserOrder(msgspec.Struct):
    """
    User order data from HyperLiquid API
    [
        {
            "coin": "BTC",
            "limitPx": "29792.0",
            "oid": 91490942,
            "side": "A",
            "sz": "0.0",
            "timestamp": 1681247412573
        }
    ]
    """

    coin: str  # Trading pair/symbol
    limitPx: str  # Limit price
    oid: int  # Order ID
    side: str  # Side: "A" for ask/sell, "B" for bid/buy
    sz: str  # Size/quantity
    timestamp: int  # Order timestamp in milliseconds


class HyperLiquidSpotToken(msgspec.Struct):
    """
    Spot token information from HyperLiquid API
    {
        "name": "USDC",
        "szDecimals": 8,
        "weiDecimals": 8,
        "index": 0,
        "tokenId": "0x6d1e7cde53ba9467b783cb7c530ce054",
        "isCanonical": true,
        "evmContract": null,
        "fullName": null
    }
    """

    name: str  # Token name/symbol
    szDecimals: int  # Size decimals
    weiDecimals: int  # Wei decimals
    index: int  # Token index
    tokenId: str  # Token ID
    isCanonical: bool  # Whether this is a canonical token
    evmContract: dict | None = None  # EVM contract address
    fullName: str | None = None  # Full token name


class HyperLiquidSpotUniverse(msgspec.Struct):
    """
    Spot universe information from HyperLiquid API
    {
        "name": "PURR/USDC",
        "tokens": [1, 0],
        "index": 0,
        "isCanonical": true
    }
    """

    name: str  # Trading pair name
    tokens: list[int]  # List of token indices
    index: int  # Universe index
    isCanonical: bool  # Whether this is a canonical pair


class HyperLiquidSpotMeta(msgspec.Struct):
    """
    Spot market metadata from HyperLiquid API
    {
        "tokens": [...],
        "universe": [...]
    }
    """

    tokens: list[HyperLiquidSpotToken]  # List of available tokens
    universe: list[HyperLiquidSpotUniverse]  # List of trading pairs


class HyperLiquidSpotBalance(msgspec.Struct):
    """
    Spot balance information from HyperLiquid API
    {
        "coin": "USDC",
        "token": 0,
        "hold": "0.0",
        "total": "14.625485",
        "entryNtl": "0.0"
    }
    """

    coin: str  # Coin/token symbol
    token: int  # Token index
    hold: str  # Amount on hold
    total: str  # Total balance
    entryNtl: str  # Entry notional value

    def parse_to_balance(self) -> Balance:
        return Balance(
            asset=self.coin,
            locked=Decimal(self.hold),
            free=Decimal(self.total) - Decimal(self.hold),
        )


class HyperLiquidUserSpotSummary(msgspec.Struct):
    """
    User spot summary from HyperLiquid API
    {
        "balances": [
            {
                "coin": "USDC",
                "token": 0,
                "hold": "0.0",
                "total": "14.625485",
                "entryNtl": "0.0"
            }
        ]
    }
    """

    balances: list[HyperLiquidSpotBalance]  # List of spot balances

    def parse_to_balances(self) -> list[Balance]:
        return [balance.parse_to_balance() for balance in self.balances]


class HyperLiquidWsMessageGeneral(msgspec.Struct):
    channel: str  # Channel name


class HyperLiquidWsBboLevelMsgData(msgspec.Struct):
    px: str  # Price
    sz: str  # Size
    n: int  # Number of orders


class HyperLiquidWsBboMsgData(msgspec.Struct):
    coin: str
    time: int
    bbo: list[HyperLiquidWsBboLevelMsgData]  # Best bid/ask levels


class HyperLiquidWsBboMsg(msgspec.Struct):
    """
    {
        "channel": "bbo",
        "data": {
            "coin": "PURR/USDC",
            "time": 1753455504649,
            "bbo": [
                {"px": "0.18134", "sz": "250.0", "n": 1},
                {"px": "0.18168", "sz": "750.0", "n": 1},
            ],
        },
    }
    """

    data: HyperLiquidWsBboMsgData


class HyperLiquidWsTradeDataMsg(msgspec.Struct):
    coin: str
    px: str  # Price
    sz: str  # Size
    time: int


class HyperLiquidWsTradeMsg(msgspec.Struct):
    data: list[HyperLiquidWsTradeDataMsg]


class HyperLiquidWsCandleDataMsg(msgspec.Struct):
    t: int  # Open time
    T: int  # Close time
    s: str  # Symbol
    i: HyperLiquidKlineInterval  # Interval
    o: str  # Open price
    c: str  # Close price
    h: str  # High price
    l: str  # Low price # noqa: E741
    v: str  # Volume
    n: int  # Number of trades


class HyperLiquidWsCandleMsg(msgspec.Struct):
    data: HyperLiquidWsCandleDataMsg


class HyperLiquidWsBasicOrder(msgspec.Struct, kw_only=True, omit_defaults=True):
    coin: str
    side: HyperLiquidOrderSide  # "A" for ask/sell, "B" for bid/buy
    limitPx: str
    sz: str
    oid: int
    timestamp: int
    origSz: str
    cloid: str | None = None  # Client order ID


class HyperLiquidWsOrderUpdatesMsgData(msgspec.Struct):
    order: HyperLiquidWsBasicOrder
    status: HyperLiquidOrderStatusType  # Order status
    statusTimestamp: int  # Timestamp of the status update


class HyperLiquidWsOrderUpdatesMsg(msgspec.Struct):
    data: list[HyperLiquidWsOrderUpdatesMsgData]


class HyperLiquidWsFills(msgspec.Struct):
    coin: str
    px: str
    sz: str
    side: HyperLiquidOrderSide  # "A" for ask/sell, "B" for bid/buy
    time: int
    startPosition: str
    dir: HyperLiquidFillDirection  # Direction of the fill
    closedPnl: str
    oid: int
    crossed: bool  # whether order crossed the spread (was taker)
    fee: str
    tid: int  # Trade ID
    feeToken: str


class HyperLiquidWsUserFillsMsgData(msgspec.Struct):
    fills: list[HyperLiquidWsFills] | None = None  # List of fills, if any


class HyperLiquidWsUserFillsMsg(msgspec.Struct):
    data: HyperLiquidWsUserFillsMsgData


class HyperLiquidWsApiOrderFilledStatus(msgspec.Struct):
    """WebSocket API order filled status"""

    totalSz: str
    avgPx: str
    oid: int


class HyperLiquidWsApiOrderStatus(msgspec.Struct, kw_only=True, omit_defaults=True):
    """WebSocket API order status"""

    error: str | None = None
    filled: HyperLiquidWsApiOrderFilledStatus | None = None
    resting: HyperLiquidOrderRestingStatus | None = None


class HyperLiquidWsApiOrderData(msgspec.Struct):
    """WebSocket API order response data"""

    statuses: list[HyperLiquidWsApiOrderStatus | str]


class HyperLiquidWsApiResponseData(msgspec.Struct):
    """WebSocket API response data wrapper"""

    type: str  # "order" or "cancel"
    data: HyperLiquidWsApiOrderData


class HyperLiquidWsApiPayload(msgspec.Struct):
    """WebSocket API payload"""

    status: str
    response: HyperLiquidWsApiResponseData


class HyperLiquidWsApiResponse(msgspec.Struct):
    """WebSocket API response"""

    type: str
    payload: HyperLiquidWsApiPayload


class HyperLiquidWsApiMessageData(msgspec.Struct):
    """WebSocket API message data"""

    id: int
    response: HyperLiquidWsApiResponse


class HyperLiquidWsApiGeneralMsg(msgspec.Struct):
    """General WebSocket API message structure"""

    channel: str
    data: HyperLiquidWsApiMessageData | None = None

    @property
    def is_pong(self) -> bool:
        return self.channel == "pong"

    @property
    def is_order_response(self) -> bool:
        return self.channel == "post" and self.data is not None

    @property
    def is_ping(self) -> bool:
        return self.channel == "ping"



import hmac
import hashlib
import httpx
import msgspec
from typing import Any, Dict, List
from urllib.parse import urljoin, urlencode
from decimal import Decimal
from nexustrader.base import ApiClient, RetryManager
from nexustrader.exchange.bybit.constants import (
    BybitBaseUrl,
    BybitRateLimiter,
    BybitRateLimiterSync,
)
from nexustrader.exchange.bybit.error import BybitError
from nexustrader.core.nautilius_core import hmac_signature, LiveClock
from nexustrader.exchange.bybit.schema import (
    BybitResponse,
    BybitOrderResponse,
    BybitPositionResponse,
    BybitOrderHistoryResponse,
    BybitOpenOrdersResponse,
    BybitWalletBalanceResponse,
    BybitKlineResponse,
    BybitIndexKlineResponse,
    BybitBatchOrderResponse,
    BybitTickersResponse,
)


class BybitApiClient(ApiClient):
    _limiter: BybitRateLimiter
    _limiter_sync: BybitRateLimiterSync

    def __init__(
        self,
        clock: LiveClock,
        api_key: str = None,
        secret: str = None,
        timeout: int = 10,
        testnet: bool = False,
        enable_rate_limit: bool = True,
        max_retries: int = 0,
        delay_initial_ms: int = 100,
        delay_max_ms: int = 800,
        backoff_factor: int = 2,
    ):
        """
        ### Testnet:
        `https://api-testnet.bybit.com`

        ### Mainnet:
        (both endpoints are available):
        `https://api.bybit.com`
        `https://api.bytick.com`

        ### Important:
        Netherland users: use `https://api.bybit.nl` for mainnet
        Hong Kong users: use `https://api.byhkbit.com` for mainnet
        Turkey users: use `https://api.bybit-tr.com` for mainnet
        Kazakhstan users: use `https://api.bybit.kz` for mainnet
        """

        super().__init__(
            clock=clock,
            api_key=api_key,
            secret=secret,
            timeout=timeout,
            rate_limiter=BybitRateLimiter(enable_rate_limit),
            rate_limiter_sync=BybitRateLimiterSync(enable_rate_limit),
            retry_manager=RetryManager(
                max_retries=max_retries,
                delay_initial_ms=delay_initial_ms,
                delay_max_ms=delay_max_ms,
                backoff_factor=backoff_factor,
                exc_types=(BybitError,),
                retry_check=lambda e: int(e.code)
                in [
                    429,
                    10000,
                    10016,
                    3400214,
                    170007,
                    177002,
                    500001,
                ],  # please refer to https://bybit-exchange.github.io/docs/v5/error
            ),
        )
        self._recv_window = 5000

        if testnet:
            self._base_url = BybitBaseUrl.TESTNET.base_url
        else:
            self._base_url = BybitBaseUrl.MAINNET_1.base_url

        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": "TradingBot/1.0",
        }

        if api_key:
            self._headers["X-BAPI-API-KEY"] = api_key

        self._msg_decoder = msgspec.json.Decoder()
        self._msg_encoder = msgspec.json.Encoder()
        self._response_decoder = msgspec.json.Decoder(BybitResponse)
        self._order_response_decoder = msgspec.json.Decoder(BybitOrderResponse)
        self._position_response_decoder = msgspec.json.Decoder(BybitPositionResponse)
        self._order_history_response_decoder = msgspec.json.Decoder(
            BybitOrderHistoryResponse
        )
        self._open_orders_response_decoder = msgspec.json.Decoder(
            BybitOpenOrdersResponse
        )
        self._wallet_balance_response_decoder = msgspec.json.Decoder(
            BybitWalletBalanceResponse
        )
        self._kline_response_decoder = msgspec.json.Decoder(BybitKlineResponse)
        self._index_kline_response_decoder = msgspec.json.Decoder(
            BybitIndexKlineResponse
        )
        self._batch_order_response_decoder = msgspec.json.Decoder(
            BybitBatchOrderResponse
        )
        self._tickers_response_decoder = msgspec.json.Decoder(BybitTickersResponse)

    def _generate_signature(self, payload: str) -> List[str]:
        timestamp = str(self._clock.timestamp_ms())

        param = str(timestamp) + self._api_key + str(self._recv_window) + payload
        hash = hmac.new(
            bytes(self._secret, "utf-8"), param.encode("utf-8"), hashlib.sha256
        )
        signature = hash.hexdigest()
        return [signature, timestamp]

    def _generate_signature_v2(self, payload: str) -> List[str]:
        timestamp = str(self._clock.timestamp_ms())
        param = f"{timestamp}{self._api_key}{self._recv_window}{payload}"
        signature = hmac_signature(self._secret, param)  # return hex digest string
        return [signature, timestamp]

    async def _fetch(
        self,
        method: str,
        base_url: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        return await self._retry_manager.run(
            name=f"{method} {endpoint}",
            func=self._fetch_async,
            method=method,
            base_url=base_url,
            endpoint=endpoint,
            payload=payload,
            signed=signed,
        )

    async def _fetch_async(
        self,
        method: str,
        base_url: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        self._init_session()

        url = urljoin(base_url, endpoint)
        payload = payload or {}

        payload_str = (
            urlencode(payload)
            if method == "GET"
            else self._msg_encoder.encode(payload).decode("utf-8")
        )

        headers = self._headers
        if signed:
            signature, timestamp = self._generate_signature_v2(payload_str)
            headers = {
                **headers,
                "X-BAPI-TIMESTAMP": timestamp,
                "X-BAPI-SIGN": signature,
                "X-BAPI-RECV-WINDOW": str(self._recv_window),
            }

        if method == "GET":
            url += f"?{payload_str}"
            payload_str = None

        try:
            self._log.debug(f"Request: {url} {payload_str}")
            response = await self._session.request(
                method=method,
                url=url,
                headers=headers,
                data=payload_str,
            )
            raw = response.content
            if response.status_code >= 400:
                raise BybitError(
                    code=response.status_code,
                    message=self._msg_decoder.decode(raw) if raw else None,
                )
            bybit_response: BybitResponse = self._response_decoder.decode(raw)
            if bybit_response.retCode == 0:
                return raw
            else:
                raise BybitError(
                    code=bybit_response.retCode,
                    message=bybit_response.retMsg,
                )
        except httpx.TimeoutException as e:
            self._log.error(f"Timeout {method} Url: {url} - {e}")
            raise
        except httpx.ConnectError as e:
            self._log.error(f"Connection Error {method} Url: {url} - {e}")
            raise
        except httpx.HTTPStatusError as e:
            self._log.error(f"HTTP Error {method} Url: {url} - {e}")
            raise
        except httpx.RequestError as e:
            self._log.error(f"Request Error {method} Url: {url} - {e}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} Url: {url} - {e}")
            raise

    def _fetch_sync(
        self,
        method: str,
        base_url: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        self._init_sync_session()

        url = urljoin(base_url, endpoint)
        payload = payload or {}

        payload_str = (
            urlencode(payload)
            if method == "GET"
            else self._msg_encoder.encode(payload).decode("utf-8")
        )

        headers = self._headers
        if signed:
            signature, timestamp = self._generate_signature_v2(payload_str)
            headers = {
                **headers,
                "X-BAPI-TIMESTAMP": timestamp,
                "X-BAPI-SIGN": signature,
                "X-BAPI-RECV-WINDOW": str(self._recv_window),
            }

        if method == "GET":
            url += f"?{payload_str}"
            payload_str = None

        try:
            self._log.debug(f"Request: {url} {payload_str}")
            response = self._sync_session.request(
                method=method,
                url=url,
                headers=headers,
                data=payload_str,
            )
            raw = response.content
            if response.status_code >= 400:
                raise BybitError(
                    code=response.status_code,
                    message=self._msg_decoder.decode(raw) if raw else None,
                )
            bybit_response: BybitResponse = self._response_decoder.decode(raw)
            if bybit_response.retCode == 0:
                return raw
            else:
                raise BybitError(
                    code=bybit_response.retCode,
                    message=bybit_response.retMsg,
                )
        except httpx.TimeoutException as e:
            self._log.error(f"Timeout {method} Url: {url} - {e}")
            raise
        except httpx.ConnectError as e:
            self._log.error(f"Connection Error {method} Url: {url} - {e}")
            raise
        except httpx.HTTPStatusError as e:
            self._log.error(f"HTTP Error {method} Url: {url} - {e}")
            raise
        except httpx.RequestError as e:
            self._log.error(f"Request Error {method} Url: {url} - {e}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} Url: {url} - {e}")
            raise

    async def post_v5_order_create(
        self,
        category: str,
        symbol: str,
        side: str,
        order_type: str,
        qty: Decimal,
        **kwargs,
    ) -> BybitOrderResponse:
        """
        https://bybit-exchange.github.io/docs/v5/order/create-order
        """
        endpoint = "/v5/order/create"
        payload = {
            "category": category,
            "symbol": symbol,
            "side": side,
            "orderType": order_type,
            "qty": str(qty),
            **kwargs,
        }
        if category == "spot":
            cost = 1
        else:
            cost = 2
        await self._limiter("trade").limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._order_response_decoder.decode(raw)

    async def post_v5_order_cancel(
        self, category: str, symbol: str, **kwargs
    ) -> BybitOrderResponse:
        """
        https://bybit-exchange.github.io/docs/v5/order/cancel-order
        """
        endpoint = "/v5/order/cancel"
        payload = {
            "category": category,
            "symbol": symbol,
            **kwargs,
        }
        if category == "spot":
            cost = 1
        else:
            cost = 2
        await self._limiter("trade").limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._order_response_decoder.decode(raw)

    def get_v5_position_list(self, category: str, **kwargs) -> BybitPositionResponse:
        endpoint = "/v5/position/list"
        payload = {
            "category": category,
            **kwargs,
        }
        self._limiter_sync("position").limit(key=endpoint, cost=1)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=True)
        return self._position_response_decoder.decode(raw)

    async def get_v5_order_realtime(self, category: str, **kwargs):
        """
        https://bybit-exchange.github.io/docs/v5/order/open-order
        """
        endpoint = "/v5/order/realtime"
        payload = {
            "category": category,
            **kwargs,
        }
        raw = await self._fetch("GET", self._base_url, endpoint, payload, signed=True)
        return self._open_orders_response_decoder.decode(raw)

    async def get_v5_order_history(self, category: str, **kwargs):
        """
        https://bybit-exchange.github.io/docs/v5/order/order-list
        """
        endpoint = "/v5/order/history"
        payload = {
            "category": category,
            **kwargs,
        }
        raw = await self._fetch("GET", self._base_url, endpoint, payload, signed=True)
        return self._order_history_response_decoder.decode(raw)

    def get_v5_account_wallet_balance(
        self, account_type: str, **kwargs
    ) -> BybitWalletBalanceResponse:
        endpoint = "/v5/account/wallet-balance"
        payload = {
            "accountType": account_type,
            **kwargs,
        }
        self._limiter_sync("account").limit(key=endpoint, cost=1)
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=True)
        return self._wallet_balance_response_decoder.decode(raw)

    async def post_v5_order_amend(
        self,
        category: str,
        symbol: str,
        orderId: str | None = None,
        orderLinkId: str | None = None,
        orderIv: str | None = None,
        triggerPrice: str | None = None,
        qty: str | None = None,
        price: str | None = None,
        tpslMode: str | None = None,
        takeProfit: str | None = None,
        stopLoss: str | None = None,
        tpTriggerBy: str | None = None,
        slTriggerBy: str | None = None,
        triggerBy: str | None = None,
        tpLimitPrice: str | None = None,
        slLimitPrice: str | None = None,
    ):
        """
        https://bybit-exchange.github.io/docs/v5/order/amend-order
        """
        endpoint = "/v5/order/amend"
        payload = {
            "category": category,
            "symbol": symbol,
            "orderId": orderId,
            "orderLinkId": orderLinkId,
            "orderIv": orderIv,
            "triggerPrice": triggerPrice,
            "qty": qty,
            "price": price,
            "tpslMode": tpslMode,
            "takeProfit": takeProfit,
            "stopLoss": stopLoss,
            "tpTriggerBy": tpTriggerBy,
            "slTriggerBy": slTriggerBy,
            "triggerBy": triggerBy,
            "tpLimitPrice": tpLimitPrice,
            "slLimitPrice": slLimitPrice,
        }
        await self._limiter("trade").limit(key=endpoint, cost=2)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._order_response_decoder.decode(raw)

    async def post_v5_order_cancel_all(
        self,
        category: str,
        symbol: str | None = None,
        baseCoin: str | None = None,
        settleCoin: str | None = None,
        orderFilter: str | None = None,
        stopOrderType: str | None = None,
    ):
        endpoint = "/v5/order/cancel-all"
        payload = {
            "category": category,
            "symbol": symbol,
            "baseCoin": baseCoin,
            "settleCoin": settleCoin,
            "orderFilter": orderFilter,
            "stopOrderType": stopOrderType,
        }
        if category == "spot":
            cost = 1
        else:
            cost = 2
        await self._limiter("trade").limit(key=endpoint, cost=cost)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._msg_decoder.decode(raw)

    def get_v5_market_kline(
        self,
        category: str,
        symbol: str,
        interval: str,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ):
        endpoint = "/v5/market/kline"
        payload = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }
        self._limiter_sync("public").limit(key=endpoint, cost=1)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=False)
        return self._kline_response_decoder.decode(raw)

    def get_v5_market_index_price_kline(
        self,
        category: str,
        symbol: str,
        interval: str,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ):
        endpoint = "/v5/market/index-price-kline"
        payload = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }
        self._limiter_sync("public").limit(key=endpoint, cost=1)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=False)
        return self._index_kline_response_decoder.decode(raw)

    async def post_v5_order_create_batch(
        self,
        category: str,
        request: List[Dict[str, Any]],
    ) -> BybitBatchOrderResponse:
        """
        Place multiple orders in a single request
        https://bybit-exchange.github.io/docs/v5/order/batch-place

        Maximum orders per request:
        - Option: 20 orders
        - Inverse: 20 orders
        - Linear: 20 orders
        - Spot: 10 orders
        """
        endpoint = "/v5/order/create-batch"
        payload = {
            "category": category,
            "request": request,
        }

        # Rate limit cost based on category
        if category == "spot":
            cost = len(request)  # 1 for each order in spot category
        else:
            cost = len(request) * 2  # 2 for each order in non-spot categories
        await self._limiter("trade").limit(key=endpoint, cost=cost)

        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._batch_order_response_decoder.decode(raw)

    def get_v5_market_tickers(
        self,
        category: str,
        symbol: str | None = None,
        base_coin: str | None = None,
        exp_date: str | None = None,
    ) -> BybitTickersResponse:
        """
        GET /v5/market/tickers

        Query for the latest price snapshot, best bid/ask price, and trading volume
        in the last 24 hours.

        Covers: Spot / USDT contract / USDC contract / Inverse contract / Option

        Args:
            category: Product type (spot, linear, inverse, option)
            symbol: Symbol name (optional), like BTCUSDT, uppercase only
            base_coin: Base coin (optional), uppercase only. Apply to option only
            exp_date: Expiry date (optional), e.g., 25DEC22. Apply to option only

        Returns:
            BybitTickersResponse: Response containing ticker data
        """
        endpoint = "/v5/market/tickers"
        payload = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "expDate": exp_date,
        }
        payload = {k: v for k, v in payload.items() if v is not None}

        self._limiter_sync("public").limit(key=endpoint, cost=1)
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=False)
        return self._tickers_response_decoder.decode(raw)



from collections import defaultdict
from typing import Dict
import re
from nexustrader.schema import (
    BookL1,
    BookL2,
    Kline,
    Trade,
    IndexPrice,
    FundingRate,
    MarkPrice,
)
from nexustrader.core.nautilius_core import MessageBus
from nexustrader.constants import KlineInterval


def _validate_indicator_name(name: str) -> bool:
    """
    Validate indicator name for use with getattr access pattern.

    Rules:
    - Must be a valid Python identifier
    - Cannot start with underscore (reserved for private attributes)
    - Cannot be a Python keyword
    - Must contain only alphanumeric characters and underscores
    - Must start with a letter or underscore (but not underscore per rule above)

    Args:
        name: The indicator name to validate

    Returns:
        bool: True if name is valid, False otherwise
    """
    import keyword

    if not name:
        return False

    # Check if it's a valid Python identifier
    if not name.isidentifier():
        return False

    # Cannot start with underscore (reserved for internal use)
    if name.startswith("_"):
        return False

    # Cannot be a Python keyword
    if keyword.iskeyword(name):
        return False

    # Additional check: must match pattern for safe attribute access
    if not re.match(r"^[a-zA-Z][a-zA-Z0-9_]*$", name):
        return False

    return True


class Indicator:
    def __init__(
        self,
        params: dict | None = None,
        name: str | None = None,
        warmup_period: int | None = None,
        warmup_interval: KlineInterval | None = None,
    ):
        indicator_name = name or type(self).__name__

        # Validate the indicator name for getattr access
        if not _validate_indicator_name(indicator_name):
            raise ValueError(
                f"Invalid indicator name '{indicator_name}'. "
                f"Name must be a valid Python identifier, cannot start with underscore, "
                f"cannot be a Python keyword, and must contain only alphanumeric characters and underscores."
            )

        self.name = indicator_name
        self.params = params
        self.warmup_period = warmup_period
        self.warmup_interval = warmup_interval
        self._is_warmed_up = False
        self._warmup_data_count = 0

    def handle_bookl1(self, bookl1: BookL1):
        raise NotImplementedError

    def handle_bookl2(self, bookl2: BookL2):
        raise NotImplementedError

    def handle_kline(self, kline: Kline):
        raise NotImplementedError

    def handle_trade(self, trade: Trade):
        raise NotImplementedError

    def handle_index_price(self, index_price: IndexPrice):
        """Handle index price updates if applicable."""
        raise NotImplementedError

    def handle_funding_rate(self, funding_rate: FundingRate):
        """Handle funding rate updates if applicable."""
        raise NotImplementedError

    def handle_mark_price(self, mark_price: MarkPrice):
        """Handle mark price updates if applicable."""
        raise NotImplementedError

    @property
    def requires_warmup(self) -> bool:
        """Check if this indicator requires warmup."""
        return self.warmup_period is not None and self.warmup_interval is not None

    @property
    def is_warmed_up(self) -> bool:
        """Check if the indicator has completed its warmup period."""
        if not self.requires_warmup:
            return True
        return self._is_warmed_up

    def _process_warmup_kline(self, kline: Kline):
        """Process a kline during warmup period."""
        if not self.requires_warmup or self._is_warmed_up:
            return

        self._warmup_data_count += 1
        self.handle_kline(kline)

        if self._warmup_data_count >= self.warmup_period:
            self._is_warmed_up = True

    def reset_warmup(self):
        """Reset warmup state. Useful for backtesting or restarting indicators."""
        self._is_warmed_up = False
        self._warmup_data_count = 0

    @property
    def value(self):
        """Get the current value of the indicator."""
        raise NotImplementedError(
            "Subclasses must implement the 'value' property to return the indicator's value."
        )


class IndicatorManager:
    def __init__(self, msgbus: MessageBus):
        self._bookl1_indicators: dict[str, list[Indicator]] = defaultdict(list)
        self._bookl2_indicators: dict[str, list[Indicator]] = defaultdict(list)
        self._kline_indicators: dict[str, list[Indicator]] = defaultdict(list)
        self._trade_indicators: dict[str, list[Indicator]] = defaultdict(list)
        self._index_price_indicators: dict[str, list[Indicator]] = defaultdict(list)
        self._funding_rate_indicators: dict[str, list[Indicator]] = defaultdict(list)
        self._mark_price_indicators: dict[str, list[Indicator]] = defaultdict(list)
        self._warmup_pending: dict[str, list[Indicator]] = defaultdict(list)

        msgbus.subscribe(topic="bookl1", handler=self.on_bookl1)
        msgbus.subscribe(topic="bookl2", handler=self.on_bookl2)
        msgbus.subscribe(topic="kline", handler=self.on_kline)
        msgbus.subscribe(topic="trade", handler=self.on_trade)
        msgbus.subscribe(topic="index_price", handler=self.on_index_price)
        msgbus.subscribe(topic="funding_rate", handler=self.on_funding_rate)
        msgbus.subscribe(topic="mark_price", handler=self.on_mark_price)

    def add_bookl1_indicator(self, symbol: str, indicator: Indicator):
        self._bookl1_indicators[symbol].append(indicator)

    def add_bookl2_indicator(self, symbol: str, indicator: Indicator):
        self._bookl2_indicators[symbol].append(indicator)

    def add_kline_indicator(self, symbol: str, indicator: Indicator):
        if indicator.requires_warmup:
            self._warmup_pending[symbol].append(indicator)
        else:
            self._kline_indicators[symbol].append(indicator)

    def add_trade_indicator(self, symbol: str, indicator: Indicator):
        self._trade_indicators[symbol].append(indicator)

    def add_index_price_indicator(self, symbol: str, indicator: Indicator):
        self._index_price_indicators[symbol].append(indicator)

    def add_funding_rate_indicator(self, symbol: str, indicator: Indicator):
        self._funding_rate_indicators[symbol].append(indicator)

    def add_mark_price_indicator(self, symbol: str, indicator: Indicator):
        self._mark_price_indicators[symbol].append(indicator)

    def on_bookl1(self, bookl1: BookL1):
        symbol = bookl1.symbol
        for indicator in self._bookl1_indicators[symbol]:
            indicator.handle_bookl1(bookl1)

    def on_bookl2(self, bookl2: BookL2):
        symbol = bookl2.symbol
        for indicator in self._bookl2_indicators[symbol]:
            indicator.handle_bookl2(bookl2)

    def on_kline(self, kline: Kline):
        symbol = kline.symbol
        for indicator in self._kline_indicators[symbol]:
            indicator.handle_kline(kline)

        # Process warmup indicators and check if they're ready
        warmup_indicators = self._warmup_pending[symbol][:]
        for indicator in warmup_indicators:
            indicator._process_warmup_kline(kline)
            if indicator.is_warmed_up:
                self._warmup_pending[symbol].remove(indicator)
                self._kline_indicators[symbol].append(indicator)

    def on_trade(self, trade: Trade):
        symbol = trade.symbol
        for indicator in self._trade_indicators[symbol]:
            indicator.handle_trade(trade)

    def on_index_price(self, index_price: IndexPrice):
        symbol = index_price.symbol
        for indicator in self._index_price_indicators[symbol]:
            indicator.handle_index_price(index_price)

    def on_funding_rate(self, funding_rate: FundingRate):
        symbol = funding_rate.symbol
        for indicator in self._funding_rate_indicators[symbol]:
            indicator.handle_funding_rate(funding_rate)

    def on_mark_price(self, mark_price: MarkPrice):
        symbol = mark_price.symbol
        for indicator in self._mark_price_indicators[symbol]:
            indicator.handle_mark_price(mark_price)

    @property
    def bookl1_subscribed_symbols(self):
        return list(self._bookl1_indicators.keys())

    @property
    def bookl2_subscribed_symbols(self):
        return list(self._bookl2_indicators.keys())

    @property
    def kline_subscribed_symbols(self):
        return list(self._kline_indicators.keys())

    @property
    def trade_subscribed_symbols(self):
        return list(self._trade_indicators.keys())

    @property
    def index_price_subscribed_symbols(self):
        return list(self._index_price_indicators.keys())

    @property
    def funding_rate_subscribed_symbols(self):
        return list(self._funding_rate_indicators.keys())

    @property
    def mark_price_subscribed_symbols(self):
        return list(self._mark_price_indicators.keys())

    def get_warmup_requirements(
        self,
    ) -> dict[str, list[tuple[Indicator, int, KlineInterval]]]:
        """Get warmup requirements for all pending indicators by symbol."""
        requirements = defaultdict(list)
        for symbol, indicators in self._warmup_pending.items():
            for indicator in indicators:
                if indicator.requires_warmup:
                    requirements[symbol].append(
                        (indicator, indicator.warmup_period, indicator.warmup_interval)
                    )
        return dict(requirements)

    def has_warmup_pending(self, symbol: str = None) -> bool:
        """Check if there are indicators pending warmup."""
        if symbol:
            return len(self._warmup_pending[symbol]) > 0
        return any(len(indicators) > 0 for indicators in self._warmup_pending.values())

    def warmup_pending_symbols(self) -> list[str]:
        """Get list of symbols with indicators pending warmup."""
        return [
            symbol for symbol, indicators in self._warmup_pending.items() if indicators
        ]


class IndicatorContainer:
    """Container for accessing indicators by symbol."""

    def __init__(self):
        self._indicators: Dict[str, Indicator] = {}

    def __getitem__(self, symbol: str):
        """Get indicator for specific symbol."""
        return self._indicators.get(symbol)

    def __setitem__(self, symbol: str, indicator: Indicator):
        """Set indicator for specific symbol."""
        self._indicators[symbol] = indicator

    def __contains__(self, symbol: str):
        """Check if symbol has an indicator."""
        return symbol in self._indicators

    def get(self, symbol: str, default=None):
        """Get indicator for symbol with default value."""
        return self._indicators.get(symbol, default)

    def symbols(self):
        """Get all symbols with indicators."""
        return list(self._indicators.keys())


class IndicatorProxy:
    """Proxy for accessing per-symbol indicator instances."""

    def __init__(self):
        self._containers: Dict[str, IndicatorContainer] = {}

    def __getattr__(self, name):
        """Get indicator container by name."""
        if name not in self._containers:
            self._containers[name] = IndicatorContainer()
        return self._containers[name]

    def register_indicator(self, name: str, symbol: str, indicator: Indicator):
        """Register an indicator instance for a specific symbol."""
        container = getattr(self, name)
        container[symbol] = indicator
